# -*- coding: utf-8 -*-
"""
Created on Sat Feb 22 16:09:28 2020
@author: Alec Neal

Last Updated: 12/16/2024
Last Editor: Kyle Koeller

Collection of functions, coefficients and equations commonly used
with short-period variable stars, but many can be used more
generally
"""
import numpy as np
from pathlib import Path
import sys
import matplotlib.pyplot as plt
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter, AutoMinorLocator, AutoLocator)
import pandas as pd


def isNaN(num):
    """
    Checks if a value is nan

    :param num: value to be checked
    :return: Boolean True or False
    """

    return num != num


def new_list(a):
    """
    Converts lists into number format with minimal decimal places

    :param a: list
    :return: new list with floats
    """
    b = []
    for i in a:
        b.append(float(format(i, ".2f")))
    return b


def conversion(a):
    """
    Converts decimal RA and DEC to standard output with colons

    :param a: decimal RA or DEC
    :return: truncated version using colons
    """
    b = []

    for i in a:
        num1 = float(i)
        if num1 < 0:
            num2 = abs((num1 - int(num1)) * 60)
            num3 = format((num2 - int(num2)) * 60, ".3f")
            num4 = float(num3)
            b.append(str(int(num1)) + ":" + str(int(num2)) + ":" + str(abs(num4)))
        else:
            num2 = (num1 - int(num1)) * 60
            num3 = format((num2 - int(num2)) * 60, ".3f")
            b.append(str(int(num1)) + ":" + str(int(num2)) + ":" + str(num3))

    return b


def splitter(a):
    """
    Splits the truncated colon RA and DEC from simbad into decimal forms

    :param a:
    :return:
    """
    # makes the coordinate string into a decimal number from the text file
    step = []
    final = []
    for i in a:
        new = i.split(":")
        num1 = int(new[0])
        if num1 < 0:
            num2 = -int(new[1])
            num3 = -int(float(new[2]))
            b = num1 + ((num2 + (num3 / 60)) / 60)
        else:
            num2 = int(new[1])
            num3 = int(float(new[2]))
            b = num1 + ((num2 + (num3 / 60)) / 60)
        step.append(format(b, ".7f"))

    for i in step:
        final.append(float(format(i)))
    return final


def decimal_limit(a):
    """
    Limits the amount of decimals that get written out for further code clarity and ease of code use

    :param a: magnitude
    :return: reduced decimal magnitude
    """
    b = list()
    for i in a:
        num = float(i)
        num2 = format(num, ".2f")
        b.append(num2)
    return b


class io:
    """
    This class provides file input/output functionalities.
    """
    def importFile_pd(inputFile, delimit=None, header=None, file_type='text', engine='python', delim_whitespace=True):
        """
        Imports a file and returns a list of columns.

        Parameters:
        - inputFile: Path to the input file.
        - delimit: Delimiter used in the file. If None, delimiter_whitespace is used for text files.
        - header: Row number(s) to use as the column names.
        - file_type: Type of the input file. Can be 'text' or 'excel'.
        - engine: Parser engine to use. Default is 'python'.
        - delim_whitespace: If True, whitespace is used as the delimiter for text files.

        Returns:
        - columnlist: A list of columns extracted from the file.
        """

        if file_type == 'text':
            if delim_whitespace:
                # Read text file with whitespace delimiter
                file = pd.read_csv(inputFile, delim_whitespace=True, header=header, engine='python')
            else:
                # Read text file with custom delimiter
                file = pd.read_csv(inputFile, sep=delimit, header=header, engine='python')
        elif file_type == 'excel':
            # Read Excel file
            file = pd.read_excel(inputFile, sep=delimit, header=header, engine='python')
        else:
            # Print error message for unsupported file types
            print('File type not currently supported. Choose text or excel type.')

        # Extract columns from the file and append them to columnlist
        columnlist = []
        for column in range(len(list(file))):
            columnlist.append(list(file[column]))

        return columnlist

    def validate_file_path(prompt):
        """
        Prompt the user to enter a file path and validate its existence.

        Parameters
        ----------
        prompt : str
            The message to display when prompting the user for input.

        Returns
        -------
        Path
            A valid Path object representing the entered file path.
        """
        while True:
            path = input(prompt)

            # Exit the program if the user types "Close"
            if path.lower() == "close":
                sys.exit()

            # Check if the entered path is valid
            try:
                images_path = Path(path)
                if images_path.exists():
                    return images_path
                else:
                    raise FileNotFoundError
            except FileNotFoundError:
                print("File not found. Please try again.")

    def validate_directory_path(prompt):
        """
        Prompt the user to enter a directory path and validate its existence.

        Parameters
        ----------
        prompt : str
            The message to display when prompting the user for input.

        Returns
        -------
        Path
            A valid Path object representing the entered directory path.
        """
        while True:
            path = input(prompt)

            # Exit the program if the user types "Close"
            if path.lower() == "close":
                sys.exit()

            # Check if the entered path is a valid directory
            try:
                directory_path = Path(path)
                if directory_path.is_dir():
                    return directory_path
                else:
                    raise NotADirectoryError
            except (FileNotFoundError, NotADirectoryError):
                print("Invalid directory. Please try again.")



# =======================================
class calc:  # assortment of functions
    def frac(x):
        """
        Returns the fractional part of a number.

        Parameters:
            x (float): The input number.

        Returns:
            float: The fractional part of the input number.
        """
        return x - np.floor(x)

    def Newton(f, x0, e=1e-8, fprime=None, max_iter=None, dx=1e-8, central_diff=True, print_iters=False):
        """
        Newton-Raphson method for finding roots of a function.

        Parameters:
            f (function): The function whose root is being found.
            x0 (float): Initial guess for the root.
            e (float): Tolerance for convergence.
            fprime (function): The derivative of the function f.
            max_iter (int): Maximum number of iterations allowed.
            dx (float): Step size for numerical differentiation.
            central_diff (bool): Whether to use central differencing for numerical differentiation.
            print_iters (bool): Whether to print the number of iterations.

        Returns:
            float or bool: The estimated root of the function, or False if maximum iterations reached.
        """
        x = x0
        iters = 0
        if fprime is None:
            if not central_diff:
                fprime = lambda x, dx=dx, func=f: (func(x + dx) / dx)

            else:
                fprime = lambda x, dx=dx, func=f: (func(x + dx) - func(x - dx)) / (2 * dx)
        if max_iter is None:
            while abs(f(x)) > e:
                x -= f(x) / fprime(x)
                iters += 1
            if print_iters:
                print('Iterations:', iters)
            return x
        else:
            iters = 0
            while abs(f(x)) > e:
                x -= f(x) / fprime(x)
                iters += 1
                if iters == max_iter:
                    break
            if iters == max_iter:
                return False
            else:
                return x

    class poly:
        def result(coeflist, value, deriv=False):
            """
            Calculates the result of a polynomial given its coefficients and a value.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the polynomial.
                value (float): The value at which the polynomial is evaluated.
                deriv (bool, optional): If True, returns the derivative of the polynomial.

            Returns:
                float: Result of the polynomial at the given value.
            """
            if deriv:
                # Derivative of the polynomial
                return sum(
                    np.array(coeflist[1:]) * np.arange(1, len(coeflist)) * value ** (np.arange(len(coeflist)) - 1))
            else:
                # Evaluate the polynomial
                return sum(np.array(coeflist) * value ** np.arange(len(coeflist)))

        def error(coeflist, value, error):
            """
            Calculates the propagated uncertainty of a polynomial.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the polynomial.
                value (float): The value at which the uncertainty is calculated.
                error (float): The error in the value.

            Returns:
                float: Propagated uncertainty of the polynomial.
            """
            errlist = [n * coef * value ** (n - 1) for n, coef in enumerate(coeflist[1:], start=1)]
            return error * sum(errlist)

        def polylist(coeflist, xmin, xmax, resolution):
            """
            Generates a list of predicted values from a polynomial within specified bounds.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the polynomial.
                xmin (float): Minimum value of the domain.
                xmax (float): Maximum value of the domain.
                resolution (int): Number of points to generate between xmin and xmax.

            Returns:
                tuple: Tuple containing x values and corresponding y values.
            """
            xlist = np.arange(xmin, xmax, (xmax - xmin) / resolution)
            ylist = [calc.poly.result(coeflist, x) for x in xlist]
            return xlist, ylist

        def regr_polyfit(x, y, deg, func=lambda x, n: x ** n, sig_y=None):
            """
            Performs a least squares polynomial fit.

            Parameters:
                x (array_like): Independent variable data.
                y (array_like): Dependent variable data.
                deg (int): Degree of the polynomial.
                func (function, optional): Function to generate features for the polynomial fit.
                sig_y (array_like, optional): Errors associated with dependent variable data.

            Returns:
                tuple: Tuple containing polynomial coefficients, standard errors, coefficient of determination,
                    predicted values, and the original model.
            """
            import statsmodels.api as sm
            x = np.array(x)
            Xlist = [func(x, n) for n in range(1, deg + 1)]
            Xstack = np.column_stack(Xlist)
            Xstack = sm.add_constant(Xstack)
            if sig_y is None:
                ogmodel = sm.OLS(y, Xstack)
            else:
                ogmodel = sm.WLS(y, Xstack, weights=1 / np.array(sig_y) ** 2)
            model = ogmodel.fit()
            return model.params, model.bse, model.rsquared, model.predict(), ogmodel

        def power(coeflist, value, base=10):
            """
            Calculates the result of a power polynomial at a given value.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the power polynomial.
                value (float): The value at which the polynomial is evaluated.
                base (float, optional): Base value of the power polynomial.

            Returns:
                float: Result of the power polynomial at the given value.
            """
            return base ** calc.poly.result(coeflist, value)

        def error_power(coeflist, value, error, base=10):
            """
            Calculates the propagated uncertainty of a power polynomial.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the power polynomial.
                value (float): The value at which the uncertainty is calculated.
                error (float): The error in the value.
                base (float, optional): Base value of the power polynomial.

            Returns:
                float: Propagated uncertainty of the power polynomial.
            """
            return abs(calc.poly.error(coeflist, value, error) * np.log(base) * calc.poly.power(coeflist, value, base))

        def t_eff_err(coeflist, value, error, temp, coeferror=[], base=10):
            """
            Calculates the effective temperature uncertainty.

            Parameters:
                coeflist (list): List of coefficients in ascending order of the polynomial.
                value (float): The value at which the uncertainty is calculated.
                error (float): The error in the value.
                temp (float): The temperature value.
                coeferror (list, optional): List of errors associated with polynomial coefficients.
                base (float, optional): Base value of the power polynomial.

            Returns:
                float: Effective temperature uncertainty.
            """
            if len(coeferror) == 0:
                return temp * np.log(base) *  calc.poly.error(coeflist, value, error)
            else:
                return temp * np.log(base) * np.sqrt( calc.poly.error(coeflist, value, error) ** 2 +
                                                     sum(np.array(coeferror) * (
                                                                 (value ** np.arange(len(coeflist))) ** 2)))

    class error:
        def per_diff(x1, x2):
            """
            Calculates the percentage difference between two values.

            Parameters:
                x1 (float): First value.
                x2 (float): Second value.

            Returns:
                float: Percentage difference between x1 and x2.
            """
            return 100 * (abs(x1 - x2) / np.mean([x1, x2]))

        def SS_residuals(obslist, modellist):
            """
            Calculates the sum of squares of residuals.

            Parameters:
                obslist (array_like): List of observed values.
                modellist (array_like): List of modeled values.

            Returns:
                float: Sum of squares of residuals.
            """
            SS_res = 0
            for n in range(len(obslist)):
                SS_res += (obslist[n] - modellist[n]) ** 2
            return SS_res

        def sig_sum(errorlist):
            """
            Calculates the sum of squared errors.

            Parameters:
                errorlist (array_like): List of errors.

            Returns:
                float: Square root of the sum of squared errors.
            """
            SS = 0
            for n in range(len(errorlist)):
                SS += errorlist[n] ** 2
            return np.sqrt(SS)

        def SS_total(obslist):
            """
            Calculates the total sum of squares.

            Parameters:
                obslist (array_like): List of observed values.

            Returns:
                float: Total sum of squares.
            """
            mean = np.mean(obslist)
            SS_tot = 0
            for n in range(len(obslist)):
                SS_tot += (obslist[n] - mean) ** 2
            return SS_tot

        def CoD(obslist, modellist):
            """
            Calculates the coefficient of determination (R^2).

            Parameters:
                obslist (array_like): List of observed values.
                modellist (array_like): List of modeled values.

            Returns:
                float: Coefficient of determination (R^2).
            """
            return 1 - calc.error.SS_residuals(obslist, modellist) / calc.error.SS_total(obslist)

        def weighted_average(valuelist, errorlist):
            """
            Calculates the weighted average and its uncertainty.

            Parameters:
                valuelist (array_like): List of values.
                errorlist (array_like): List of corresponding errors.

            Returns:
                tuple: Weighted average value, uncertainty in the average, and total weight.
            """
            M = sum(1 / np.array(errorlist) ** 2)
            w_average = 0
            for n in range(len(errorlist)):
                w_average += valuelist[n] / errorlist[n] ** 2
            w_average /= M
            ave_error = 1 / np.sqrt(M)
            return w_average, ave_error, M

        def avg(errorlist):
            """
            Calculates the average of a list of errors.

            Parameters:
                errorlist (array_like): List of errors.

            Returns:
                float: Average error.
            """
            error2list = [error ** 2 for error in errorlist]
            return np.sqrt(sum(error2list)) * (1 / len(errorlist))

        def red_X2(obslist, modellist, obserror):
            """
            Calculates the reduced chi squared.

            Parameters:
                obslist (array_like): List of observed values.
                modellist (array_like): List of modeled values.
                obserror (array_like): List of observed errors.

            Returns:
                float: Reduced chi squared value.
            """
            X2v0 = 0
            X2v = X2v0
            for n in range(len(obslist)):
                X2v += ((obslist[n] - modellist[n]) / obserror[n]) ** 2
            return X2v

        def truncnorm(size, lower=-3.0, upper=3.0, mean=0.0, sigma=1.0):
            """
            Generates truncated Gaussian deviates.

            Parameters:
                size (int): Number of deviates to generate.
                lower (float, optional): Lower bound of the truncated range.
                upper (float, optional): Upper bound of the truncated range.
                mean (float, optional): Mean of the Gaussian distribution.
                sigma (float, optional): Standard deviation of the Gaussian distribution.

            Returns:
                array_like: Array of truncated Gaussian deviates.
            """
            import scipy.stats as sci
            return sci.truncnorm.rvs(lower, upper, loc=mean, scale=sigma, size=size)

    # -------------------
    class astro:
        class convert:
            def HJD_phase(HJDlist, period, Epoch, Pdot=0):
                """
                Converts a list of Heliocentric Julian dates to phase values.

                Parameters:
                    HJDlist (array_like): List of Heliocentric Julian dates.
                    period (float): Period of the cyclic phenomenon.
                    Epoch (float): Reference epoch for phase calculation.
                    Pdot (float, optional): Rate of change of period. Default is 0.

                Returns:
                    array_like: List of corresponding phase values.
                """
                daydiff = np.array(HJDlist) - Epoch
                return (daydiff / (period + Pdot * daydiff)) - np.floor(daydiff / (period + Pdot * daydiff))

            def JD_to_Greg(JD):
                """
                Converts Julian date to Gregorian date format.

                Parameters:
                    JD (float): Julian date.

                Returns:
                    str: Gregorian date in the format 'YYYY MM DD'.
                """
                # months = ['none', 'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September',
                #           'October', 'November', 'December']
                f = JD + 1401 + int((int((4 * JD + 274277) / 146097) * 3) / 4) - 38
                e = 4 * f + 3
                g = int((e % 1461) / 4)
                h = 5 * g + 2
                D = int((h % 153) / 5) + 1
                M = (int(h / 153) + 2) % 12 + 1
                Y = int(e / 1461) - 4716 + int((12 + 2 - M) / 12)
                if len(str(D)) < 2:
                    D = '0' + str(D)
                if len(str(M)) < 2:
                    M = '0' + str(M)
                return str(Y) + ' ' + str(M) + ' ' + str(D)

            class magToflux:
                """
                Converts magnitude to flux values.
                """

                def flux(mag):
                    """
                    Converts magnitude to flux.

                    Parameters:
                        mag (float): Magnitude value.

                    Returns:
                        float: Flux value.
                    """
                    return 10 ** (-0.4 * mag)

                def error(mag, magerr):
                    """
                    Calculates flux error from magnitude error.

                    Parameters:
                        mag (float): Magnitude value.
                        magerr (float): Magnitude error value.

                    Returns:
                        float: Flux error value.
                    """
                    return 0.4 * magerr * np.log(10) * 10 ** (-0.4 * mag)

            class fluxTomag:
                """
                Converts flux to magnitude values.
                """

                def mag(flux):
                    """
                    Converts flux to magnitude.

                    Parameters:
                        flux (float): Flux value.

                    Returns:
                        float: Magnitude value.
                    """
                    return -2.5 * np.log10(flux)

                def error(flux, fluxerr):
                    """
                    Calculates magnitude error from flux error.

                    Parameters:
                        flux (float): Flux value.
                        fluxerr (float): Flux error value.

                    Returns:
                        float: Magnitude error value.
                    """
                    return (2.5 * fluxerr) / (flux * np.log(10))


# ======================================
class binning:
    def makebin(phase, bins, phasefluxlist):
        """
        Creates bins and groups phase and flux data into them.

        Parameters:
            phase (float): Phase value to bin around.
            bins (int): Number of bins.
            phasefluxlist (list): List of phase and flux data.

        Returns:
            tuple: Two lists containing phase and flux values grouped into bins.
        """
        halfbin = 0.5 / bins
        binphases = []
        binfluxes = []
        n0 = 0
        n = n0
        while (n < len(phasefluxlist)):
            if phase == 0:
                if 0 < phasefluxlist[n] < halfbin:
                    binphases.append(phasefluxlist[n])
                    binfluxes.append(phasefluxlist[n + 1])
                elif 1 - halfbin < phasefluxlist[n] < 1:
                    binphases.append(phasefluxlist[n])
                    binfluxes.append(phasefluxlist[n + 1])
            else:
                if phase - halfbin < phasefluxlist[n] < phase + halfbin:
                    binphases.append(phasefluxlist[n])
                    binfluxes.append(phasefluxlist[n + 1])
            n += 2
        return binphases, binfluxes

    def binall(bins, phasefluxlist):
        """
        Bins all phases and their corresponding fluxes.

        Parameters:
            bins (int): Number of bins.
            phasefluxlist (list): List of phase and flux data.

        Returns:
            tuple: Lists containing binned phases and averaged binned fluxes.
        """
        binnedphaselist = []
        binnedfluxlist = []
        dphase = 1 / bins
        phase0 = 0
        phase = phase0
        while (phase < 1):
            binnedphaselist.append(phase)
            binnedfluxlist.append(np.mean(binning.makebin(phase, bins, phasefluxlist)[1]))
            phase += dphase
        return binnedphaselist, binnedfluxlist

    def norm_flux(binnedfluxlist, ob_fluxlist, ob_fluxerr, norm_factor='bin'):
        """
        Normalize flux lists based on specified normalization factor.

        Parameters:
            binnedfluxlist (list): List of binned flux values.
            ob_fluxlist (list): List of observed flux values.
            ob_fluxerr (list): List of errors associated with observed flux values.
            norm_factor (str): Normalization factor, either 'bin' or 'ob'. Defaults to 'bin'.

        Returns:
            tuple: Three lists containing normalized binned fluxes, observed fluxes, and their errors.
        """
        if norm_factor == 'ob':
            normf = max(ob_fluxlist)
        else:
            normf = max(binnedfluxlist)
        norm_binned = [binnedflux / normf for binnedflux in binnedfluxlist]
        norm_ob = [ob_flux / normf for ob_flux in ob_fluxlist]
        norm_err = [ob_err / normf for ob_err in ob_fluxerr]
        return norm_binned, norm_ob, norm_err

    def minibinner(phaselist, fluxlist, bins):
        """
        Returns binned lists given a list of phases and fluxes, and specified number of bins.

        Parameters:
            phaselist (list): List of phases.
            fluxlist (list): List of fluxes.
            bins (int): Number of bins.

        Returns:
            tuple: Four lists containing binned phases, binned fluxes, normalized binned fluxes, and normalized observed fluxes.
        """
        obs = len(phaselist)
        phaseflux = [(phaselist[n], fluxlist[n]) for n in range(obs)]
        binned = binning.binall(bins, phaseflux)
        binnedphaselist = binned[0]
        binnedfluxlist = binned[1]
        normed = binning.norm_flux(binnedfluxlist, fluxlist, fluxlist)
        n_binnedfluxlist = normed[0]
        n_ob_fluxlist = normed[1]
        return binnedphaselist, binnedfluxlist, n_binnedfluxlist, n_ob_fluxlist

    def masterbinner(HJD, mag, magerr, Epoch, period, bins=40, weighted=True, norm_factor='alt', centered=True, pdot=0):
        # Convert HJD to phases
        ob_phaselist = calc.astro.convert.HJD_phase(HJD, period, Epoch, Pdot=pdot)

        # Extract observed magnitudes and errors
        ob_maglist = mag
        ob_magerr = magerr
        observations = len(ob_maglist)

        # Convert magnitudes to fluxes
        ob_fluxlist = list(calc.astro.convert.magToflux.flux(np.array(ob_maglist)))
        ob_fluxerr = list(calc.astro.convert.magToflux.error(np.array(ob_maglist), np.array(ob_magerr)))

        # Define bin properties
        halfbin = 0.5 / bins
        dphase = 1 / bins

        # Define function to create bins
        def makebin(phase):
            phases_in_bin = []
            fluxes_in_bin = []
            errors_in_bin = []
            index_in_bin = []
            for n in range(observations):
                if centered == False:
                    if phase < ob_phaselist[n] < phase + dphase:
                        phases_in_bin.append(ob_phaselist[n])
                        fluxes_in_bin.append(ob_fluxlist[n])
                        errors_in_bin.append(ob_fluxerr[n])
                        index_in_bin.append(n)
                else:
                    if phase == 0:
                        if 0 < ob_phaselist[n] < halfbin:
                            phases_in_bin.append(ob_phaselist[n])
                            fluxes_in_bin.append(ob_fluxlist[n])
                            errors_in_bin.append(ob_fluxerr[n])
                            index_in_bin.append(n)
                        elif 1 - halfbin < ob_phaselist[n] < 1:
                            phases_in_bin.append(ob_phaselist[n])
                            fluxes_in_bin.append(ob_fluxlist[n])
                            errors_in_bin.append(ob_fluxerr[n])
                            index_in_bin.append(n)
                    else:
                        if phase - halfbin < ob_phaselist[n] < phase + halfbin:
                            phases_in_bin.append(ob_phaselist[n])
                            fluxes_in_bin.append(ob_fluxlist[n])
                            errors_in_bin.append(ob_fluxerr[n])
                            index_in_bin.append(n)
            return phases_in_bin, fluxes_in_bin, errors_in_bin, index_in_bin

        # Initialize lists for binned data
        binnedfluxlist = []
        binnederrorlist = []
        avgphaselist = []
        master_phases_in_bin = []
        master_fluxes_in_bin = []
        master_errors_in_bin = []
        master_index_in_bin = []

        # Generate binned data
        binnedphaselist = np.arange(0, 1, dphase)
        for phase in binnedphaselist:
            stuff_at_phase = makebin(phase)
            phases_at_phase = stuff_at_phase[0]
            fluxes_at_phase = stuff_at_phase[1]
            errors_at_phase = stuff_at_phase[2]
            index_at_phase = stuff_at_phase[3]
            master_phases_in_bin.append(np.array(phases_at_phase))
            master_fluxes_in_bin.append(np.array(fluxes_at_phase))
            master_errors_in_bin.append(np.array(errors_at_phase))
            master_index_in_bin.append(index_at_phase)
            if weighted == False:
                binnedfluxlist.append(np.mean(fluxes_at_phase))
                binnederrorlist.append(calc.error.avg(errors_at_phase))
            else:
                w_average = calc.error.weighted_average(fluxes_at_phase, errors_at_phase)
                binnedfluxlist.append(w_average[0])
                binnederrorlist.append(w_average[1])
            avgphaselist.append(np.mean(phases_at_phase))

        # Convert binned fluxes to magnitudes
        binnedmaglist = list(calc.astro.convert.fluxTomag.mag(np.array(binnedfluxlist)))
        binnedmagerr = list(calc.astro.convert.fluxTomag.error(np.array(binnedfluxlist), np.array(binnederrorlist)))

        # Determine normalization factor
        if norm_factor == 'obs':
            norm_f = max(ob_fluxlist)
        elif norm_factor == 'bin':
            norm_f = max(binnedfluxlist)
        elif norm_factor == 'avgmag':
            norm_f = 10 ** (-0.4 * np.mean(ob_maglist))
        else:
            offset = 0.025
            quad1 = []
            quad2 = []
            for n in range(len(ob_phaselist)):
                if 0.25 - offset < ob_phaselist[n] < 0.25 + offset:
                    quad1.append(ob_fluxlist[n])
                elif 0.75 - offset < ob_phaselist[n] < 0.75 + offset:
                    quad2.append(ob_fluxlist[n])
            norm_f = max([np.mean(quad1), np.mean(quad2)])

        # Normalize lists
        def normlist(valuelist, norm_f):
            return np.array(valuelist) / norm_f

        n_binnedfluxlist = normlist(binnedfluxlist, norm_f)
        n_ob_fluxlist = normlist(ob_fluxlist, norm_f)
        n_ob_fluxerr = normlist(ob_fluxerr, norm_f)
        n_binnederrorlist = normlist(binnederrorlist, norm_f)
        for n in range(len(master_fluxes_in_bin)):
            master_fluxes_in_bin[n] /= norm_f
            master_errors_in_bin[n] /= norm_f

        # Assemble data into lists
        master_time = [binnedphaselist, ob_phaselist, avgphaselist, HJD]
        master_norm = [n_binnedfluxlist, n_ob_fluxlist, n_ob_fluxerr, n_binnederrorlist]
        master_ob_flux = [binnedfluxlist, ob_fluxlist, ob_fluxerr, binnederrorlist]
        master_ob_mag = [binnedmaglist, ob_maglist, ob_magerr, binnedmagerr]
        master_stuff_in_bin = [master_phases_in_bin, master_fluxes_in_bin, master_errors_in_bin, master_index_in_bin]

        return master_time, master_norm, master_ob_flux, master_ob_mag, norm_f, master_stuff_in_bin

    def masterbinner_FF(fileName, Epoch, period, bins,
                        HJDcol=0, magcol=1, magerrcol=2,
                        weighted=True, norm_factor='bin', sep=None,
                        header=None, file_type='text', centered=True, pdot=0):
        # Import data from file
        file = io.importFile_pd(fileName, delimit=sep, header=header, file_type=file_type)

        # Extract columns from the file
        HJD = file[HJDcol]
        mag = file[magcol]
        magerr = file[magerrcol]

        # Apply masterbinner function to the extracted data
        MB = binning.masterbinner(HJD, mag, magerr, Epoch, period, bins,
                                  weighted=weighted, norm_factor=norm_factor, centered=centered, pdot=pdot)

        return MB

    def minipolybinner(c_master_phases, c_master_fluxes, nc_master_phases, nc_master_fluxes,
                       section_order, section_res=None):
        # Calculate the number of sections
        sections = len(c_master_phases)

        # Determine section resolution if not provided
        if section_res is None:
            section_res = int(128 / sections)

        # Initialize lists to store polynomial phase and flux
        section_polyphase = []
        section_polyflux = []
        lastphase = []
        lastflux = []

        # Calculate parameters
        halfsec = 0.5 / sections
        dphase = 1 / sections
        bound1 = int(section_res * 0.25)
        bound2 = int(section_res * 0.75)

        # Iterate over each section
        for section in range(sections):
            if section == 0:
                # Handle the first section separately to account for phase wrapping
                for n in range(len(c_master_phases[0])):
                    if 1 - halfsec < c_master_phases[0][n] < 1:
                        c_master_phases[0][n] -= 1
                # Fit polynomial to the section's phase and flux data
                c_coef = calc.poly.regr_polyfit(c_master_phases[0], c_master_fluxes[0], section_order)[0]
                c_polylist = calc.poly.polylist(c_coef, -halfsec, halfsec, section_res)
                c_polyphase = c_polylist[0][bound1:bound2:]
                c_polyflux = c_polylist[1][bound1:bound2:]
                # Split the polynomial data into appropriate sections
                for n in range(len(c_polyphase)):
                    if c_polyphase[n] < 0:
                        lastphase.append(c_polyphase[n] + 1)
                        lastflux.append(c_polyflux[n])
                    else:
                        section_polyphase.append(c_polyphase[n])
                        section_polyflux.append(c_polyflux[n])
                # Fit polynomial to the neighboring section's phase and flux data
                nc_coef = calc.poly.regr_polyfit(nc_master_phases[0], nc_master_fluxes[0], section_order)[0]
                nc_polylist = calc.poly.polylist(nc_coef, 0, dphase, section_res)
                section_polyphase += list(nc_polylist[0][bound1:bound2:])
                section_polyflux += list(nc_polylist[1][bound1:bound2:])
            else:
                # Fit polynomial to the current section's phase and flux data
                c_coef = calc.poly.regr_polyfit(c_master_phases[section], c_master_fluxes[section], section_order)[0]
                c_polylist = calc.poly.polylist(c_coef, section / sections - halfsec, section / sections + halfsec,
                                                section_res)
                section_polyphase += list(c_polylist[0][bound1:bound2:])
                section_polyflux += list(c_polylist[1][bound1:bound2:])
                # Fit polynomial to the neighboring section's phase and flux data
                nc_coef = calc.poly.regr_polyfit(nc_master_phases[section], nc_master_fluxes[section], section_order)[0]
                nc_polylist = calc.poly.polylist(nc_coef, section / sections, section / sections + dphase, section_res)
                section_polyphase += list(nc_polylist[0][bound1:bound2:])
                section_polyflux += list(nc_polylist[1][bound1:bound2:])

        # Add the data from the last phase
        section_polyphase += list(lastphase)
        section_polyflux += list(lastflux)

        # Check if the first phase value is zero
        if section_polyphase[0] != 0.0:
            print('WARNING: Unfit for FT, first phase value not zero.')

        return section_polyphase, section_polyflux

    def polybinner(input_file, Epoch, period, sections=4, norm_factor='alt',
                   section_order=8, FT_order=12, section_res=None, HJD_mag_magerr=[],
                   mag_coef=False, pdot=0):
        # Set section resolution if not provided
        if section_res == None:
            section_res = int(128 / sections)

        # Perform master binning based on provided data or input file
        if len(HJD_mag_magerr) == 3:
            c_MB = binning.masterbinner(HJD_mag_magerr[0], HJD_mag_magerr[1], HJD_mag_magerr[2], Epoch, period,
                                        sections, centered=True, norm_factor=norm_factor, pdot=pdot)
            nc_MB = binning.masterbinner(HJD_mag_magerr[0], HJD_mag_magerr[1], HJD_mag_magerr[2], Epoch, period,
                                         sections, centered=False, norm_factor=norm_factor, pdot=pdot)
        else:
            c_MB = binning.masterbinner_FF(input_file, Epoch, period, sections, centered=True, norm_factor=norm_factor,
                                           pdot=pdot)
            nc_MB = binning.masterbinner_FF(input_file, Epoch, period, sections, centered=False,
                                            norm_factor=norm_factor, pdot=pdot)

        # Extract master binning data
        c_master_phases = c_MB[5][0]
        c_master_fluxes = c_MB[5][1]
        nc_master_phases = nc_MB[5][0]
        nc_master_fluxes = nc_MB[5][1]

        # ob_phaselist = c_MB[0][1]
        # ob_fluxlist = c_MB[1][1]

        # Initialize lists to store polynomial phase and flux
        section_polyphase = []
        section_polyflux = []
        lastphase = []
        lastflux = []
        halfsec = 0.5 / sections
        dphase = 1 / sections
        bound1 = int(section_res * 0.25)
        bound2 = int(section_res * 0.75)

        # Iterate over each section
        for section in range(sections):
            if section == 0:
                # Handle the first section separately to account for phase wrapping
                for n in range(len(c_master_phases[0])):
                    if 1 - halfsec < c_master_phases[0][n] < 1:
                        c_master_phases[0][n] -= 1
                c_coef = calc.poly.regr_polyfit(c_master_phases[0], c_master_fluxes[0], section_order)[0]
                c_polylist = calc.poly.polylist(c_coef, -halfsec, halfsec, section_res)
                c_polyphase = c_polylist[0][bound1:bound2:]
                c_polyflux = c_polylist[1][bound1:bound2:]
                for n in range(len(c_polyphase)):
                    if c_polyphase[n] < 0:
                        lastphase.append(c_polyphase[n] + 1)
                        lastflux.append(c_polyflux[n])
                    else:
                        section_polyphase.append(c_polyphase[n])
                        section_polyflux.append(c_polyflux[n])
                nc_coef = calc.poly.regr_polyfit(nc_master_phases[0], nc_master_fluxes[0], section_order)[0]
                nc_polylist = calc.poly.polylist(nc_coef, 0, dphase, section_res)
                section_polyphase += list(nc_polylist[0][bound1:bound2:])
                section_polyflux += list(nc_polylist[1][bound1:bound2:])
            else:
                c_coef = calc.poly.regr_polyfit(c_master_phases[section], c_master_fluxes[section], section_order)[0]
                c_polylist = calc.poly.polylist(c_coef, section / sections - halfsec, section / sections + halfsec,
                                                section_res)
                section_polyphase += list(c_polylist[0][bound1:bound2:])
                section_polyflux += list(c_polylist[1][bound1:bound2:])

                nc_coef = calc.poly.regr_polyfit(nc_master_phases[section], nc_master_fluxes[section], section_order)[0]
                nc_polylist = calc.poly.polylist(nc_coef, section / sections, section / sections + dphase, section_res)
                section_polyphase += list(nc_polylist[0][bound1:bound2:])
                section_polyflux += list(nc_polylist[1][bound1:bound2:])

        section_polyphase += list(lastphase)
        section_polyflux += list(lastflux)

        # Calculate Fourier transform coefficients
        if mag_coef == True:
            a, b = FT.coefficients(-2.5 * np.log10(np.array(section_polyflux) * c_MB[4]))[1:3]
        else:
            a, b = FT.coefficients(section_polyflux)[1:3]

        return [a, b], [c_MB, nc_MB], [section_polyphase, section_polyflux]


# ======================================
class FT:  # Fourier transform
    def coefficients(binnedvaluelist):
        """
        Calculates Fourier coefficients based on numPy's fft.ifft function

        Parameters
        ----------
        binnedvaluelist : array like
            List of evenly spaced (binned) values

        Returns
        -------
        Fbfli : array
            combines both a & b coefficients, ignore

        coslist : array
            cosine FT coefficients (a)

        sinlist : array
            sine FT coefficients (b)

        """
        Fbfli = np.fft.ifft(binnedvaluelist, norm=None) * 2
        coslist = np.real(Fbfli)
        sinlist = np.imag(Fbfli)
        coslist[0] = np.mean(binnedvaluelist)
        sinlist[0] = 0
        return Fbfli, coslist, sinlist

    def a_sig_fast(a, b, term, aterm, ob_phase, ob_flux, ob_fluxerr, order, dx0=1):
        """
        Estimates the significance of an 'a' coefficient in Fourier fitting.

        Parameters:
            a (list): List of 'a' coefficients.
            b (list): List of 'b' coefficients.
            term (int): Index of the 'a' coefficient being evaluated.
            aterm (float): Initial guess for the value of the 'a' coefficient.
            ob_phase (array-like): Array of observed phases.
            ob_flux (array-like): Array of observed fluxes.
            ob_fluxerr (array-like): Array of observed flux errors.
            order (int): Order of the Fourier fit.
            dx0 (float): Step size for numerical differentiation (default is 1).

        Returns:
            tuple: A tuple containing the estimated significance and the ratio of upper to lower bounds.
        """
        C = calc.error.red_X2(ob_flux, FT.synth(a, b, ob_phase, order), ob_fluxerr) + 1

        def spec_a(mod_a):
            a[term] = mod_a
            F = calc.error.red_X2(ob_flux, FT.synth(a, b, ob_phase, order), ob_fluxerr) - C
            return F

        upper = abs(calc.Newton(spec_a, aterm + dx0, 1e-5, dx=1e-10) - aterm)
        lower = abs(calc.Newton(spec_a, aterm - dx0, 1e-5, dx=1e-10) - aterm)
        a[term] = aterm
        return (upper + lower) / 2, upper / lower

    def b_sig_fast(a, b, term, bterm, ob_phase, ob_flux, ob_fluxerr, order, dx0=1):
        """
        Estimates the significance of a 'b' coefficient in Fourier fitting.

        Parameters:
            a (list): List of 'a' coefficients.
            b (list): List of 'b' coefficients.
            term (int): Index of the 'b' coefficient being evaluated.
            bterm (float): Initial guess for the value of the 'b' coefficient.
            ob_phase (array-like): Array of observed phases.
            ob_flux (array-like): Array of observed fluxes.
            ob_fluxerr (array-like): Array of observed flux errors.
            order (int): Order of the Fourier fit.
            dx0 (float): Step size for numerical differentiation (default is 1).

        Returns:
            tuple: A tuple containing the estimated significance and the ratio of upper to lower bounds.
        """
        C = calc.error.red_X2(ob_flux, FT.synth(a, b, ob_phase, order), ob_fluxerr) + 1

        def spec_b(mod_b):
            b[term] = mod_b
            F = calc.error.red_X2(ob_flux, FT.synth(a, b, ob_phase, order), ob_fluxerr) - C
            return F

        upper = abs(calc.Newton(spec_b, bterm + dx0, 1e-5, dx=1e-10) - bterm)
        lower = abs(calc.Newton(spec_b, bterm - dx0, 1e-5, dx=1e-10, central_diff=True) - bterm)
        b[term] = bterm
        return (upper + lower) / 2, upper / lower

    def sumatphase(phase, order, a, b):
        """
        Calculates the amplitude (value) of the Fourier Transform at a given phase.

        Parameters:
            phase (float): Phase at which to calculate the amplitude.
            order (int): Order of the Fourier fit.
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.

        Returns:
            float: Amplitude of the Fourier Transform at the given phase.
        """
        orders = np.arange(order + 1)
        return sum(
            a[:order + 1:] * np.cos(2 * np.pi * phase * orders) + b[:order + 1:] * np.sin(2 * np.pi * phase * orders))

    def deriv_sumatphase(phase, order, a, b):
        """
        Calculates the derivative of the Fourier Transform at a given phase.

        Parameters:
            phase (float): Phase at which to calculate the derivative.
            order (int): Order of the Fourier fit.
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.

        Returns:
            float: Derivative of the Fourier Transform at the given phase.
        """
        deriv_atphase = 0
        for k in range(1, order + 1):
            deriv_atphase += (-b[k] * np.sin(2 * np.pi * k * phase) - a[k] * np.cos(
                2 * np.pi * k * phase)) * 4 * np.pi ** 2 * k ** 2
        return deriv_atphase

    def unc_sumatphase(phase, order, a_unc, b_unc):
        """
        Calculates the uncertainty of the Fourier Transform at a given phase,
        given lists of 'a' and 'b' uncertainties.

        Parameters:
            phase (float): Phase at which to calculate the uncertainty.
            order (int): Order of the Fourier fit.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.

        Returns:
            float: Uncertainty of the Fourier Transform at the given phase.
        """
        unc_I = a_unc[0] ** 2
        for k in range(1, order + 1):
            unc_I += (a_unc[k] * np.cos(2 * np.pi * k * phase)) ** 2 + (b_unc[k] * np.sin(2 * np.pi * k * phase)) ** 2
        return np.sqrt(unc_I)

    def FT_plotlist(a, b, order, resolution):
        """
        Generates the results of a Fourier Transform given the coefficients.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.
            resolution (int): Number of points in the resulting plot.

        Returns:
            tuple: A tuple containing lists of phases, Fourier Transform values, and derivative values.
        """
        phase = 0
        FTfluxlist = []
        FTphaselist = []
        FTderivlist = []
        while (phase < 1):
            FTfluxlist.append(FT.sumatphase(phase, order, a, b))
            FTderivlist.append(FT.deriv_sumatphase(phase, order, a, b))
            FTphaselist.append(round(phase, 7))
            phase += (1 / resolution)
        return FTphaselist, FTfluxlist, FTderivlist

    def sim_ob_flux(FT, ob_fluxerr, lower=-3.0, upper=3.0):
        """
        Modifies a pre-generated synthetic Fourier light-curve amplitude list,
        and alters each value by a capped 3 sigma Gaussian deviate of the observational
        error given the index in ob_fluxerr. The two lists should be sorted by the same
        phase, but sim_FT_curve does this naturally.

        This function is mainly for the sim_FT_curve program and not a standalone function.

        Parameters:
            FT (array-like): Synthetic Fourier light-curve amplitude list.
            ob_fluxerr (array-like): Array of observational flux errors.
            lower (float): Lower limit for Gaussian deviate (default is -3.0).
            upper (float): Upper limit for Gaussian deviate (default is 3.0).

        Returns:
            array-like: Modified synthetic Fourier light-curve amplitude list.
        """
        obs = len(ob_fluxerr)
        devlist = calc.error.truncnorm(obs, lower=lower, upper=upper)
        sim_ob_flux = []
        for n in range(obs):
            sim_ob_flux.append(FT[n] + devlist[n] * ob_fluxerr[n])
        return sim_ob_flux

    def synth(a, b, ob_phaselist, order):
        """
        Generates a synthetic light curve using the Fourier coefficients.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            ob_phaselist (array-like): Array of observed phases.
            order (int): Order of the Fourier fit.

        Returns:
            array-like: Synthetic light curve.
        """
        synthlist = []
        for phase in ob_phaselist:
            synthlist.append(FT.sumatphase(phase, order, a, b))
        return synthlist

    def int_sumatphase(a, b, phase, order):
        """
        Calculates the integral of the Fourier Transform at a given phase.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            phase (float): Phase at which to calculate the integral.
            order (int): Order of the Fourier fit.

        Returns:
            float: Integral of the Fourier Transform at the given phase.
        """
        nlist = np.arange(order + 1)[1::]
        return a[0] * phase + (0.5 / np.pi) * sum((a[1:order + 1:] * np.sin(2 * np.pi * nlist * phase) - b[
                                                                                                         1:order + 1:] * np.cos(
            2 * np.pi * nlist * phase)) / nlist)

    def integral(a, b, order, lowerphase, upperphase):
        """
        Calculates the definite integral of the Fourier Transform.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.
            lowerphase (float): Lower bound of the integral.
            upperphase (float): Upper bound of the integral.

        Returns:
            float: Definite integral of the Fourier Transform.
        """
        uppersum = FT.int_sumatphase(a, b, upperphase, order)
        lowersum = FT.int_sumatphase(a, b, lowerphase, order)
        return uppersum - lowersum

    def int_unc_atphase(phase, a_unc, b_unc):
        """
        Calculates the uncertainty of the definite integral at a given phase.

        Parameters:
            phase (float): Phase at which to calculate the uncertainty.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.

        Returns:
            float: Uncertainty of the definite integral at the given phase.
        """
        orderp1 = len(a_unc)
        int_unc = 0
        for k in range(1, orderp1):
            int_unc += ((a_unc[k] / k) * np.sin(2 * np.pi * phase * k)) ** 2 + (
                    (b_unc[k] / k) * np.cos(2 * np.pi * phase * k)) ** 2
        int_unc *= 0.25 / np.pi ** 2
        int_unc += (a_unc[0] * phase) ** 2
        return np.sqrt(int_unc)


# ======================================
class OConnell:  # O'Connell effect
    def OER_FT(a, b, order):
        """
        Calculates the Odd-Even Ratio (OER) of the Fourier Transform.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            float: Odd-Even Ratio of the Fourier Transform.
        """
        nlist = np.arange(order + 1)[1::]
        A = 0.5 * sum(a[nlist])
        B = sum(b[nlist[::2]] / nlist[::2]) / np.pi
        return 1 - (2 / (1 + A / B))

    def OER_FT_error(a, b, a_unc, b_unc, order):
        """
        Calculates both the Odd-Even Ratio (OER) and its error of the Fourier Transform.
        Requires the 'a' and 'b' coefficients, and their respective uncertainties.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            tuple: A tuple containing the Odd-Even Ratio and its error.
        """
        I0_area = sum(a[:order + 1:]) * 0.5
        Top = FT.integral(a, b, order, 0, 0.5) - I0_area
        Bot = FT.integral(a, b, order, 0.5, 1) - I0_area
        OER = Top / Bot
        sigsum = calc.error.sig_sum(a_unc)
        uncTop = np.sqrt((FT.int_unc_atphase(0.5, a_unc, b_unc)) ** 2 + (FT.int_unc_atphase(0, a_unc, b_unc)) ** 2 + (
                0.5 * sigsum) ** 2)
        uncBot = np.sqrt((FT.int_unc_atphase(1, a_unc, b_unc)) ** 2 + (FT.int_unc_atphase(0.5, a_unc, b_unc)) ** 2 + (
                0.5 * sigsum) ** 2)
        OER_unc = OER * np.sqrt((uncTop / Top) ** 2 + (uncBot / Bot) ** 2)
        return OER, OER_unc

    def OER_FT_error_fixed(a, b, a_unc, b_unc, order):
        """
        Calculates the error of the Odd-Even Ratio (OER) of the Fourier Transform with fixed coefficients.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            float: Error of the Odd-Even Ratio of the Fourier Transform.
        """
        a = np.array(a)
        b = np.array(b)
        a_unc = np.array(a_unc)
        b_unc = np.array(b_unc)
        nlist = np.arange(order + 1)[1::]
        A = 0.5 * sum(a[nlist])
        B = sum(b[nlist[::2]] / nlist[::2]) / np.pi
        sig_A2 = 0.25 * sum(a_unc[nlist] ** 2)
        sig_B2 = sum((b_unc[nlist[::2]] / nlist[::2]) ** 2) / (np.pi ** 2)
        return np.sqrt(sig_A2 / A ** 2 + sig_B2 / B ** 2) / abs(1 + A / (2 * B) + B / (2 * A))

    def LCA_FT(a, b, order, resolution):
        """
        Calculates the Light Curve Asymmetry (LCA) given the 'a' and 'b' coefficients.
        Must specify the resolution for numerical integration.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.
            resolution (int): Number of points for integration.

        Returns:
            float: Light Curve Asymmetry (LCA).
        """
        import scipy
        Phi = np.linspace(0, 0.5, resolution)
        no_a = np.zeros(len(a))
        K2 = []
        for phase in Phi:
            K2.append(((2 * FT.sumatphase(phase, order, no_a, b)) / FT.sumatphase(phase, order, a, b)) ** 2)
        LCA = np.sqrt(scipy.integrate.simps(K2, Phi))
        return LCA

    def L_error(phase, order, a, b, a_unc, b_unc):
        """
        Calculates the error in the LCA integrand, which is then used to find the error in the LCA.
        Requires 'a' and 'b' coefficients and their uncertainties.

        Parameters:
            phase (float): Phase value.
            order (int): Order of the Fourier fit.
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.

        Returns:
            tuple: A tuple containing the LCA integrand value and its error.
        """
        I = FT.sumatphase(phase, order, a, b)  # Compute the flux value at the given phase
        J = 2 * FT.sumatphase(phase, order, np.zeros(order + 1), b)  # Compute a component of the LCA integrand
        K = J / I
        L = K ** 2  # Calculate the LCA integrand value

        # Calculate partial derivatives of L with respect to 'a' and 'b' coefficients
        dL_da0 = -2 * L / I
        dL_dak = []
        dL_dbk = []
        for k in range(1, order + 1):
            dL_dak.append(dL_da0 * np.cos(2 * np.pi * k * phase))
            dL_dbk.append(2 * np.sin(2 * np.pi * k * phase) * (2 * J / I ** 2 - J ** 2 / I ** 3))

        # Compute the uncertainty in the LCA integrand using error propagation
        L_err = (dL_da0 * a_unc[0]) ** 2
        for n in range(len(dL_dak)):
            L_err += (dL_dak[n] * a_unc[n + 1]) ** 2 + (dL_dbk[n] * b_unc[n + 1]) ** 2
        L_err = np.sqrt(L_err)

        return L, L_err

    def LCA_FT_error(a, b, a_unc, b_unc, order, resolution):
        """
        Calculates the uncertainty in the Fourier Transform Light Curve Asymmetry (LCA).

        This function computes the uncertainty in the LCA by integrating the error contributions
        at different phases using Simpson's rule. It utilizes the L_error function to calculate
        the LCA integrand at each phase, which provides both the LCA value and its uncertainty.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.
            resolution (int): Number of points for integration.

        Returns:
            tuple: A tuple containing the LCA value and its uncertainty.
        """
        import scipy

        # Generate phase values for integration
        Phi = np.linspace(0, 0.5, resolution)

        # Initialize lists to store LCA integrand values and their uncertainties at each phase
        Llist = []
        Lerrlist = []

        # Compute the LCA integrand and its uncertainties at each phase
        for phase in Phi:
            L_ap = OConnell.L_error(phase, order, a, b, a_unc, b_unc)
            Llist.append(L_ap[0])
            Lerrlist.append(L_ap[1])

        # Integrate the LCA integrand and its uncertainties using Simpson's rule
        int_L = scipy.integrate.simps(Llist, Phi)
        int_L_error = scipy.integrate.simps(Lerrlist, Phi)

        # Compute the LCA value and its uncertainty
        LCA = OConnell.LCA_FT(a, b, order, resolution)
        LCA_error = 0.5 * LCA * (int_L_error / int_L)

        return LCA, LCA_error

    def LCA_FT_error2(a, b, a_unc, b_unc, order, resolution):
        """
        Calculates the uncertainty in the Fourier Transform Light Curve Asymmetry (LCA) using an alternative method.

        This function computes the uncertainty in the LCA by directly evaluating the LCA integrand
        at different phases. It then integrates these values to obtain the total uncertainty in the LCA.
        It is an alternative approach to LCA_FT_error, utilizing a different method for calculating the
        uncertainty contribution at each phase.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.
            resolution (int): Number of points for integration.

        Returns:
            tuple: A tuple containing the LCA value and its uncertainty.

        """
        a, a_unc = np.array(a), np.array(a_unc)
        b, b_unc = np.array(b), np.array(b_unc)

        def L_error2(phase):
            """
            Calculates the error contribution for the Fourier Transform Light Curve Asymmetry (LCA) integrand at a given phase.

            This function computes the error contribution of the LCA integrand at a specified phase.
            It first calculates the derivative of the flux with respect to phase, 'dIphi', and the flux value 'I'
            at the given phase using the provided 'a' and 'b' coefficients. Then, it computes the LCA integrand
            error term 'L', which is the square of the ratio of 'dIphi' to 'I'. Additionally, it calculates the
            uncertainty in 'L' using error propagation, considering the uncertainties in 'a' and 'b' coefficients.

            Parameters:
                phase (float): Phase at which to compute the error contribution.

            Returns:
                tuple: A tuple containing the LCA integrand error term 'L' and its uncertainty.
            """
            nlist = np.arange(order + 1)  # Generate the list of orders
            dIphi = 2 * sum(
                b[nlist] * np.sin(2 * np.pi * nlist * phase))  # Compute the derivative of flux with respect to phase
            I = FT.sumatphase(phase, order, a, b)  # Compute the flux value at the given phase
            L = (dIphi / I) ** 2  # Calculate the LCA integrand error term
            # Calculate the uncertainty in the LCA integrand using error propagation
            L_err = 2 / I * (2 * L ** 0.5 - L) * np.sqrt(sum((b_unc * np.sin(2 * np.pi * phase * nlist)) ** 2))
            return L, L_err

        import scipy
        Phi = np.linspace(0, 0.5, resolution)
        Llist = []
        Lerrlist = []
        for phase in Phi:
            L_ap = L_error2(phase)
            Llist.append(L_ap[0])
            Lerrlist.append(L_ap[1])
        int_L = scipy.integrate.simps(Llist, Phi)
        int_L_error = scipy.integrate.simps(Lerrlist, Phi)
        # print('L_err =',int_L_error,'L =',int_L)
        LCA = OConnell.LCA_FT(a, b, order, resolution)
        LCA_error = 0.5 * LCA * (int_L_error / int_L)
        return LCA, LCA_error

    def Delta_I(a, b, order):
        """
        Directly calculates the difference in flux at the two quadratures (0.25, 0.75),
        given the 'a' and 'b' coefficients.

        This function calculates the difference in flux between two quadrature phases (0.25 and 0.75),
        providing insight into the O'Connell effect. It returns the difference in flux, as well as the
        flux values at the specified phases.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            tuple: A tuple containing the difference in flux, flux at phase 0.25, and flux at phase 0.75.
        """
        Ip = FT.sumatphase(0.25, order, a, b)
        Is = FT.sumatphase(0.75, order, a, b)
        dI = Ip - Is
        return dI, Ip, Is

    def Delta_I_error(a, b, a_unc, b_unc, order):
        """
        Calculates the uncertainty in dI of the Fourier Transform.

        Parameters:
            a (array-like): List of 'a' coefficients.
            b (array-like): List of 'b' coefficients.
            a_unc (array-like): List of uncertainties for 'a' coefficients.
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            tuple: A tuple containing the dI value and its uncertainty.
        """
        # Calculate uncertainties at quadrature phases
        sig_Ip = FT.unc_sumatphase(0.25, order, a_unc, b_unc)
        sig_Is = FT.unc_sumatphase(0.75, order, a_unc, b_unc)

        # Compute total uncertainty in dI
        sig_dI = np.sqrt(sig_Ip ** 2 + sig_Is ** 2)

        # Calculate dI value
        dI = OConnell.Delta_I(a, b, order)[0]

        return dI, sig_dI

    def Delta_I_fixed(b, order):
        """
        Calculates the difference in flux at the two quadratures (0.25, 0.75)
        for a given set of 'b' coefficients.

        Parameters:
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            float: Difference in flux at quadratures.
        """
        mlist = np.arange(int(np.ceil(order / 2)))
        return 2 * sum((-1) ** mlist * b[2 * mlist + 1])

    def Delta_I_error_fixed(b_unc, order):
        """
        Calculates the uncertainty in dI_fixed of the Fourier Transform.

        Parameters:
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.

        Returns:
            float: Uncertainty in dI_fixed.
        """
        # Calculate total uncertainty in dI_fixed
        return 2 * np.sqrt(sum(np.array(b_unc)[1:order + 1:2] ** 2))

    def dI_at_phase(b, order, phase):
        """
        Calculates the difference in flux at a specific phase for a given set of 'b' coefficients.

        Parameters:
            b (array-like): List of 'b' coefficients.
            order (int): Order of the Fourier fit.
            phase (float): Phase value.

        Returns:
            float: Difference in flux at the specified phase.
        """
        nlist = np.arange(order + 1)
        return 2 * sum(b[nlist] * np.sin(2 * np.pi * nlist * phase))

    def dI_at_phase_error(b_unc, order, phase):
        """
        Calculates the uncertainty in dI at a specific phase for a given set of 'b' coefficients.

        Parameters:
            b_unc (array-like): List of uncertainties for 'b' coefficients.
            order (int): Order of the Fourier fit.
            phase (float): Phase value.

        Returns:
            float: Uncertainty in dI at the specified phase.
        """
        nlist = np.arange(order + 1)
        return 2 * np.sqrt(sum((b_unc[nlist] * np.sin(2 * np.pi * nlist * phase)) ** 2))

    def Delta_I_mean_obs(ob_phaselist, ob_fluxlist, ob_fluxerr, phase_range=0.05, weighted=False):
        """
        Calculates the difference of flux at quadrature from observational fluxes
        and the uncertainty of this measure.

        Parameters:
            ob_phaselist (array-like): List of observational phase values.
            ob_fluxlist (array-like): List of observational flux values.
            ob_fluxerr (array-like): List of observational flux errors.
            phase_range (float): Range around quadrature phase to consider.
            weighted (bool): Flag to indicate whether to use weighted averaging.

        Returns:
            tuple: A tuple containing the mean difference in flux at quadrature and its uncertainty.
        """
        Iplist = []
        Iperrors = []
        Islist = []
        Iserrors = []

        # Extract fluxes and errors within specified phase ranges from quadrature
        for n in range(len(ob_phaselist)):
            if 0.25 - phase_range < ob_phaselist[n] < 0.25 + phase_range:
                Iplist.append(ob_fluxlist[n])
                Iperrors.append(ob_fluxerr[n])
            if 0.75 - phase_range < ob_phaselist[n] < 0.75 + phase_range:
                Islist.append(ob_fluxlist[n])
                Iserrors.append(ob_fluxerr[n])

        if weighted:
            # Calculate weighted averages if specified
            Ipmean = calc.error.weighted_average(Iplist, Iperrors)
            Ismean = calc.error.weighted_average(Islist, Iserrors)
            dI_mean_obs = Ipmean[0] - Ismean[0]
            dI_mean_error = np.sqrt(Ipmean[1] ** 2 + Ismean[1] ** 2)
        else:
            # Calculate simple means otherwise
            dI_mean_obs = np.mean(Iplist) - np.mean(Islist)
            dI_mean_error = np.sqrt(calc.error.avg(Iperrors) ** 2 + calc.error.avg(Iserrors) ** 2)

        return dI_mean_obs, dI_mean_error

    def Delta_I_mean_obs_noerror(ob_phaselist, ob_fluxlist, phase_range=0.05):
        """
        Calculates the mean difference in flux at quadrature from observational fluxes
        without considering observational errors.

        Parameters:
            ob_phaselist (array-like): List of observational phase values.
            ob_fluxlist (array-like): List of observational flux values.
            phase_range (float): Range around quadrature phase to consider.

        Returns:
            float: Mean difference in flux at quadrature from observational fluxes.
        """
        Iplist = []
        Islist = []

        # Extract fluxes within specified phase ranges from quadrature
        for n in range(len(ob_phaselist)):
            if 0.25 - phase_range < ob_phaselist[n] < 0.25 + phase_range:
                Iplist.append(ob_fluxlist[n])
            if 0.75 - phase_range < ob_phaselist[n] < 0.75 + phase_range:
                Islist.append(ob_fluxlist[n])

        # Calculate mean difference in flux at quadrature
        dI_mean_obs = np.mean(Iplist) - np.mean(Islist)

        return dI_mean_obs


# Function to calculate the sum of reciprocals of squared errors
def M(errorlist):
    """
    Calculates the sum of reciprocals of squared errors.

    Parameters:
        errorlist (array-like): List of errors.

    Returns:
        float: Sum of reciprocals of squared errors.
    """
    M0 = 0
    M = M0
    for error in errorlist:
        M += 1 / error ** 2
    return M


# Function to calculate the weight factor
def wfactor(errorlist, n, M):
    """
    Calculates the weight factor for a given error.

    Parameters:
        errorlist (array-like): List of errors.
        n (int): Index of the error in the list.
        M (float): Sum of reciprocals of squared errors.

    Returns:
        float: Weight factor for the error at index n.
    """
    return 1 / (errorlist[n] ** 2 * M)


# Class Flower containing functions for calculations related to Flower 1996 and Torres 2010
class Flower:
    class T:
        # Coefficients for Teff calculation
        c = [3.97914510671409, -0.654992268598245, 1.74069004238509, -4.60881515405716, 6.79259977994447,
             -5.39690989132252, 2.19297037652249, -0.359495739295671]

        def Teff(BV, error):
            """
            Calculates the effective temperature (Teff) using the B-V color index.

            Parameters:
                BV (float): B-V color index.
                error (float): Error in B-V color index.

            Returns:
                tuple: A tuple containing the Teff value and its error.
            """
            # Calculate Teff using polynomial approximation
            temp = calc.poly.power(Flower.T.c, BV, 10)
            # Calculate error in Teff using error propagation
            err = calc.poly.t_eff_err(Flower.T.c, BV, error, temp)
            return temp, err



# Class containing functions for calculations related to Harmanec
class Harmanec:
    class mass:
        # Coefficients for M1 calculation
        c = [-121.6782, 88.057, -21.46965, 1.771141]

        def M1(BV):
            """
            Calculates the mass of a star using the B-V color index.

            Parameters:
                BV (float): B-V color index.

            Returns:
                float: Mass of the star.
            """
            # Calculate M1 using polynomial approximation
            M1 = 10 ** (calc.poly.result(Harmanec.mass.c, np.log10(Flower.T.Teff(BV))))
            return M1


# Class containing constants and functions for interstellar reddening
class Red:
    # Constants for color excess
    J_K = 0.17084
    J_H = 0.10554
    V_R = 0.58 / 3.1

    def colorEx(filter1, filter2, Av):
        """
        Calculates the color excess in a specified filter combination.

        Parameters:
            filter1 (str): First filter.
            filter2 (str): Second filter.
            Av (float): V-band extinction.

        Returns:
            float: Color excess.
        """
        excess = str(filter1) + '_' + str(filter2)
        if excess == 'J_K':
            return Av * Red.J_K
        elif excess == 'J_H':
            return Av * Red.J_H
        elif excess == 'V_R':
            return Av * Red.V_R



# Class for plotting functions
class plot:
    def amp(valuelist):
        """
        Calculates the amplitude of a list of values.

        Parameters:
            valuelist (list): List of values.

        Returns:
            float: Amplitude of the values.
        """
        return max(valuelist) - min(valuelist)

    def aliasing2(phaselist, maglist, errorlist, alias=0.6):
        """
        Filters the phase, magnitude, and error lists to keep only data within the specified alias range.

        Parameters:
            phaselist (list): List of phase values.
            maglist (list): List of magnitude values.
            errorlist (list): List of error values.
            alias (float): Alias range.

        Returns:
            tuple: Filtered phase, magnitude, and error lists.
        """
        phase = list(np.array(phaselist) - 1) + list(phaselist)
        mag = list(maglist) + list(maglist)
        error = list(errorlist) + list(errorlist)
        a_phase = []
        a_mag = []
        a_error = []
        for n in range(len(phase)):
            if -alias < phase[n] < alias:
                a_phase.append(phase[n])
                a_mag.append(mag[n])
                a_error.append(error[n])
        return a_phase, a_mag, a_error

    def multiplot(figsize=(8, 8), dpi=256, height_ratios=[3, 1], hspace=0, sharex=True, sharey=False, fig=None):
        """
        Creates a multiplot with specified parameters.

        Parameters:
            figsize (tuple): Figure size.
            dpi (int): Dots per inch.
            height_ratios (list): List of height ratios for subplots.
            hspace (float): Height space between subplots.
            sharex (bool): Share x-axis among subplots.
            sharey (bool): Share y-axis among subplots.
            fig (object): Figure object.

        Returns:
            tuple: Axes and figure objects.
        """
        if fig == None:
            fig = plt.figure(1, figsize=figsize, dpi=dpi)
        axs = fig.subplots(len(height_ratios), sharex=sharex, sharey=sharey,
                           gridspec_kw={'hspace': hspace, 'height_ratios': height_ratios})
        return axs, fig

    def sm_format(ax, X=None, x=None, Y=None, y=None, Xsize=7, xsize=3.5, tickwidth=1,
                  xtop=True, xbottom=True, yright=True, yleft=True, numbersize=12, autoticks=True,
                  topspine=True, bottomspine=True, rightspine=True, leftspine=True, xformatter=True,
                  xdirection='in', ydirection='in', spines=True):
        """
        Formats the plot axis.

        Parameters:
            ax (object): Axis object.
            X (float): Major tick spacing on x-axis.
            x (float): Minor tick spacing on x-axis.
            Y (float): Major tick spacing on y-axis.
            y (float): Minor tick spacing on y-axis.
            Xsize (float): Size of major ticks on x-axis.
            xsize (float): Size of minor ticks on x-axis.
            tickwidth (float): Tick width.
            xtop (bool): Display ticks on top of x-axis.
            xbottom (bool): Display ticks on bottom of x-axis.
            yright (bool): Display ticks on right side of y-axis.
            yleft (bool): Display ticks on left side of y-axis.
            numbersize (int): Size of tick labels.
            autoticks (bool): Use automatic tick spacing.
            topspine (bool): Display top spine.
            bottomspine (bool): Display bottom spine.
            rightspine (bool): Display right spine.
            leftspine (bool): Display left spine.
            xformatter (bool): Use x-axis tick formatter.
            xdirection (str): Tick direction on x-axis.
            ydirection (str): Tick direction on y-axis.
            spines (bool): Display axis spines.

        Returns:
            str: 'DONE' when formatting is complete.
        """
        ax.tick_params(axis='x', which='major', length=Xsize, width=tickwidth, direction=xdirection, top=xtop,
                       bottom=xbottom, labelsize=numbersize)
        ax.tick_params(axis='y', which='major', length=Xsize, width=tickwidth, direction=ydirection, right=yright,
                       labelsize=numbersize, left=yleft)
        ax.tick_params(axis='x', which='minor', length=xsize, width=tickwidth, direction=xdirection, top=xtop,
                       bottom=xbottom)
        ax.tick_params(axis='y', which='minor', length=xsize, width=tickwidth, direction=ydirection, right=yright,
                       left=yleft)
        if spines == False:
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['left'].set_visible(False)
        else:
            ax.spines['top'].set_linewidth(tickwidth)
            ax.spines['left'].set_linewidth(tickwidth)
            ax.spines['right'].set_linewidth(tickwidth)
            ax.spines['bottom'].set_linewidth(tickwidth)
            ax.spines['bottom'].set_visible(bottomspine)
            ax.spines['top'].set_visible(topspine)
            ax.spines['right'].set_visible(rightspine)
            ax.spines['left'].set_visible(leftspine)
        if X == None:
            ax.xaxis.set_major_locator(AutoLocator())
        else:
            ax.xaxis.set_major_locator(MultipleLocator(X))
        if x == None:
            ax.xaxis.set_minor_locator(AutoMinorLocator())
        else:
            ax.xaxis.set_minor_locator(MultipleLocator(x))
        if Y == None:
            ax.yaxis.set_major_locator(AutoLocator())
        else:
            ax.yaxis.set_major_locator(MultipleLocator(Y))
        if y == None:
            ax.yaxis.set_minor_locator(AutoMinorLocator())
        else:
            ax.yaxis.set_minor_locator(MultipleLocator(y))
        if xformatter == True:
            ax.xaxis.set_major_formatter(FormatStrFormatter('$%g$'))
        return 'DONE'



# ======================================
class Roche:
    def Kopal_cyl(rho, phi, z, q):
        """
        Calculates the Kopal potential for a cylindrical coordinate system.

        Parameters:
            rho (float): Radial coordinate.
            phi (float): Azimuthal coordinate.
            z (float): Vertical coordinate.
            q (float): Mass ratio.

        Returns:
            float: Value of the Kopal potential.
        """
        return 1 / np.sqrt(rho ** 2 + z ** 2) + q / (
            np.sqrt(1 + rho ** 2 + z ** 2 - 2 * rho * np.cos(phi))) - q * rho * np.cos(phi) + 0.5 * (1 + q) * rho ** 2

    def gen_Kopal_cyl(rho, phi, z, q,
                      xcm=None, ycm=0, zcm=0,
                      potcap=None):
        """
        Generates the Kopal potential for a cylindrical coordinate system.

        Parameters:
            rho (float): Radial coordinate.
            phi (float): Azimuthal coordinate.
            z (float): Vertical coordinate.
            q (float): Mass ratio.
            xcm (float, optional): x-coordinate of the center of mass.
            ycm (float, optional): y-coordinate of the center of mass.
            zcm (float, optional): z-coordinate of the center of mass.
            potcap (float, optional): Potential cap.

        Returns:
            float: Value of the generated Kopal potential.
        """
        if xcm == None:
            xcm = q / (1 + q)
        A1 = -q / (1 + q)
        A2 = 1 / (1 + q)
        B1 = xcm ** 2 + ycm ** 2 + zcm ** 2 + 2 * xcm * A1 + A1 ** 2
        B2 = xcm ** 2 + ycm ** 2 + zcm ** 2 + 2 * xcm * A2 + A2 ** 2
        X = rho * np.cos(phi)
        Y = rho * np.sin(phi)
        s1 = np.sqrt(rho ** 2 + z ** 2 - 2 * (X * (xcm + A1) + Y * ycm + z * zcm) + B1)
        s2 = np.sqrt(rho ** 2 + z ** 2 - 2 * (X * (xcm + A2) + Y * ycm + z * zcm) + B2)
        rw2 = rho ** 2 - 2 * (xcm * X + ycm * Y) + xcm ** 2 + ycm ** 2
        potent = 1 / s1 + q / s2 + 0.5 * (1 + q) * rw2 - 0.5 * q ** 2 / (1 + q)
        return potent

    def Lagrange_123(q, e=1e-8):
        """
        Calculates the Lagrange points L1, L2, and L3 for a given mass ratio.

        Parameters:
            q (float): Mass ratio.
            e (float, optional): Tolerance for the Newton-Raphson method.

        Returns:
            tuple: Values of Lagrange points L1, L2, and L3.
        """
        L1 = lambda x: q / x ** 2 - x * (1 + q) - 1 / (1 - x) ** 2 + 1
        L2 = lambda x: q / x ** 2 - x * (1 + q) + 1 / (1 + x) ** 2 - 1
        L3 = lambda x: 1 / (q * x ** 2) - x * (1 + 1 / q) + 1 / (1 + x) ** 2 - 1
        xL1 = calc.Newton(L1, 0.5, e=e)
        xL2 = calc.Newton(L2, 0.5, e=e)
        xL3 = calc.Newton(L3, 0.5, e=e)
        return xL1, xL2, xL3

    def Kopal_zero(rho, phi, z, q, Kopal, body='M1'):
        """
        Calculates the Kopal potential for zero equilibrium.

        Parameters:
            rho (float): Radial coordinate.
            phi (float): Azimuthal coordinate.
            z (float): Vertical coordinate.
            q (float): Mass ratio.
            Kopal (float): Kopal potential.
            body (str, optional): Body type ('M1', 'M2', 'dM1', 'dM2').

        Returns:
            float: Value of the Kopal potential for zero equilibrium.
        """
        r = rho
        if body == 'M1':
            return 1 / np.sqrt(r ** 2 + z ** 2) + q / np.sqrt(
                1 + r ** 2 + z ** 2 - 2 * r * np.cos(phi)) - q * r * np.cos(phi) + 0.5 * (1 + q) * r ** 2 - Kopal
        elif body == 'M2':
            return q / np.sqrt(r ** 2 + z ** 2) + 1 / np.sqrt(1 + r ** 2 + z ** 2 + 2 * r * np.cos(phi)) + r * np.cos(
                phi) + 0.5 * (1 + q) * r ** 2 + (1 - q) / 2 - Kopal
        elif body == 'dM1':
            return -r / (r ** 2 + z ** 2) ** (3 / 2) - q * (r - np.cos(phi)) / (
                    1 + r ** 2 + z ** 2 - 2 * r * np.cos(phi)) ** (3 / 2) - q * np.cos(phi) + (1 + q) * r
        elif body == 'dM2':
            return -r * q / (r ** 2 + z ** 2) ** (3 / 2) - (r + np.cos(phi)) / (
                    1 + r ** 2 + z ** 2 + 2 * r * np.cos(phi)) ** (3 / 2) + np.cos(phi) + (1 + q) * r
        else:
            return print('Invalid body, choose M1 or M2.')

    def gen_Kopal_zero(rho, phi, z, q, Kopal,
                       xcm=None, ycm=0, zcm=0):
        """
        Generates the Kopal potential for zero equilibrium.

        Parameters:
            rho (float): Radial coordinate.
            phi (float): Azimuthal coordinate.
            z (float): Vertical coordinate.
            q (float): Mass ratio.
            Kopal (float): Kopal potential.
            xcm (float, optional): x-coordinate of the center of mass.
            ycm (float, optional): y-coordinate of the center of mass.
            zcm (float, optional): z-coordinate of the center of mass.

        Returns:
            float: Value of the generated Kopal potential for zero equilibrium.
        """
        return Roche.gen_Kopal_cyl(rho, phi, z, q, xcm=xcm, ycm=ycm, zcm=zcm) - Kopal


class Pecaut:  # V-R effective temperature fit from Pecaut and Mamajek 2013 https://arxiv.org/pdf/1307.2657.pdf
    class T:
        """
        Class T defines the effective temperature (Teff) as a function of color index (V-R).

        Attributes:
            c1 (list): Coefficients for Teff calculation when -0.115 < V-R <= 0.019 (B1V to A1V).
            c2 (list): Coefficients for Teff calculation when 0.019 < V-R < 1.079 (A1V to M3V).
            c1_err (list): Coefficients errors for c1.
            c2_err (list): Coefficients errors for c2.
        """

        c1 = [3.98007, -1.2742, 32.41373, 98.06855]  # -0.115 < V-R < 0.019
        c2 = [3.9764, -0.80319, 0.64832, -0.2651]  # 0.019 < V-R < 1.079

        c1_err = [0.00499, 0.25431, 6.60765, 41.6003]
        c2_err = [0.00179, 0.01409, 0.03136, 0.0197]

        def Teff(VR, error):
            """
            Calculates the effective temperature (Teff) given the color index (V-R).

            Parameters:
                VR (float): Color index (V-R).
                error (float): Error in color index (V-R).

            Returns:
                tuple: A tuple containing the calculated Teff value and its error.
            """
            if -0.115 < VR <= 0.019:  # B1V to A1V
                temp = calc.poly.power(Pecaut.T.c1, VR, 10)
                err = calc.poly.t_eff_err(Pecaut.T.c1, VR, error, temp, coeferror=Pecaut.T.c1_err)
                return temp, err
            elif 0.019 < VR < 1.079:  # A1V to M3V
                temp = calc.poly.power(Pecaut.T.c2, VR, 10)
                err = calc.poly.t_eff_err(Pecaut.T.c2, VR, error, temp, coeferror=Pecaut.T.c2_err)
                return temp, err
            else:
                return 0, 0




"""
Look up the TESS data and download that data onto a local drive.
Author: Kyle Koeller
Created: 2/19/2022
Last Updated: 01/25/2026
"""

# import required packages
from astroquery.mast import Tesscut
from .tesscut import process_tess_cutout
from os.path import exists
import pandas as pd
import os

from astropy import units as u
import pkg_resources


def run_tess_search(system_name, download_all, specific_sector, download_path, write_callback, cancel_event):
    """
    Search for TESS data and download the specified sectors with cancel functionality.

    Parameters
    ----------
    system_name : str
        The TIC ID or system name.
    download_all : bool
        Whether to download all available sectors.
    specific_sector : int, optional
        The specific sector to download if `download_all` is False.
    download_path : str, optional
        The directory to save downloaded files.
    write_callback : function, optional
        Callback to log progress or errors to the GUI.
    """

    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)
    try:
        # Check for cancellation
        if cancel_event.is_set():
            log("Task canceled before starting.")
            return

        # Validate system name
        log(f"Searching for sectors for system: {system_name}")
        sector_table = Tesscut.get_sectors(objectname=system_name)

        if not sector_table:
            raise ValueError(f"No TESS data found for system {system_name}.")

        # The ccd info comes from this paper:
        # https://archive.stsci.edu/files/live/sites/mast/files/home/missions-and-data/active-missions/tess/_documents/TESS_Instrument_Handbook_v0.1.pdf
        filename = pkg_resources.resource_filename(__name__, 'tess_ccd_info.txt')
        dc = pd.read_csv(filename, header=None, sep="\t", skiprows=[0])

        gain = dc[3]  # videoscale, gain for the individual camera/ccd
        tess_camera = dc[0]  # camera number
        tess_ccd = dc[1]  # ccd number
        # slice = dc[2]  # could be used in the future

        sector_camera = []
        sector_ccd = []
        for count, val in enumerate(sector_table["camera"]):
            sector_camera.append(val)
            sector_ccd.append(sector_table["ccd"][count])

        # create a tuple form of the sector and TESS camera/ccd data
        list_gain = []
        a = list(zip(tess_camera, tess_ccd))  # tess cam and ccd's
        b = list(zip(sector_camera, sector_ccd))  # sect cam and ccd's

        # compare the TESS and sector information
        for _, sect in enumerate(b):
            for y, tess in enumerate(a):
                if sect == tess:
                    list_gain.append(gain[y])

        # splits up list_gain into lists by every 1, 2, 3, or 4th value since there are 4 slices for each camera and ccd
        A = list_gain[::4]
        B = list_gain[1::4]
        C = list_gain[2::4]
        D = list_gain[3::4]

        # append the gain values to the sector table list
        sector_table.add_column(A, name="A gain")
        sector_table.add_column(B, name="B gain")
        sector_table.add_column(C, name="C gain")
        sector_table.add_column(D, name="D gain")

        # prints off the sector table to let the user know what sectors TESS has observed the object
        log("\nThe read noise for the cameras is between 7-11 electrons/pixel.")

        # Log the sector table
        formatted_table = "\n".join(sector_table.pformat(show_name=True, max_width=-1, align="^"))
        log("Sector Table:\n" + formatted_table)

        # Check specific_sector logic
        if not download_all:
            log("Downloading all available sectors.")
            for sector in sector_table["sector"]:
                if cancel_event.is_set():
                    log(f"Task canceled while processing Sector {sector}.")
                    return
                download_sector(system_name, sector, download_path, write_callback, cancel_event)
        else:
            if specific_sector:
                log(f"Downloading specific sector: {specific_sector}.")
                download_sector(system_name, specific_sector, download_path, write_callback, cancel_event)
            else:
                log("Error: Specific sector is not specified.")
                raise ValueError("Specific sector is not specified.")

        log("Finished downloading all sector data related to " + system_name + "\n")
    except Exception as e:
        log(f"An error occurred during TESS Database Search: {e}")


def download_sector(system_name, sector, download_path, write_callback, cancel_event):
    """
    Download TESS sector data for a given system.

    Parameters
    ----------
    system_name : str
        The TIC ID or system name.
    sector : int
        The sector number to download.
    download_path : str
        The path to save downloaded data.
    write_callback: function
        to write log messages to the GUI
    """

    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    try:
        if not exists(download_path):
            raise FileNotFoundError(f"The path '{download_path}' does not exist.")

        # Create sector-specific directory if it doesn't exist
        sector_path = os.path.join(download_path, str(sector))
        os.makedirs(sector_path, exist_ok=True)  # Create the directory if it doesn't exist
        log(f"Directory created or already exists: {sector_path}")

        log(f"Starting download for Sector {sector}.")
        manifest = Tesscut.download_cutouts(
            objectname=system_name, size=[30, 30] * u.arcmin, sector=sector, path=sector_path)
        process_tess_cutout(
            search_file=manifest,  # Replace with actual file
            pathway=sector_path,
            sector=sector,
            outprefix=f"{system_name}_S{sector}_",
            write_callback=write_callback,
            cancel_event=cancel_event
        )
        log(f"Completed download for Sector {sector}.")
    except Exception as e:
        log(f"Failed to download Sector {sector}: {e}")
        raise




# -*- coding: utf-8 -*-
"""
Calculates the O'Connel Effect based on this paper: https://app.aavso.org/jaavso/article/3511/

Created on Thu Feb 25 00:47:37 2021
Last Edited: 06/18/2025

Original Author: Alec Neal
Last Edits Done By: Kyle Koeller
"""

# Importing necessary libraries
import matplotlib.pyplot as plt
from .vseq_updated import plot, binning, calc, FT, OConnell
from tqdm import tqdm
import numpy as np
import statistics as st
from os import path

# Lambda function to calculate sigma of a function
sig_f = lambda f, x, sig_x: abs(f(x + sig_x) - f(x - sig_x)) / 2


# Main function for calculating O'Connell Effect
def main(filepath="", filter_files=None, obj_name="", period=0, hjd=0, write_callback=None, cancel_event=None):
    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    try:
        if cancel_event.is_set():
            log("Task canceled.")
            return

        multi_OConnell_total(filter_files, hjd, period, order=10, sims=1000,
                             sections=4, section_order=7, plot_only=False, save=True,
                             outName=(filepath + "\\" + obj_name + ".pdf"),
                             write_callback=write_callback, cancel_event=cancel_event)
    except Exception as e:
        log(f"An error occurred: {e}")
        raise


def quick_tex(thing):
    """
    Quick TeX formatting function.
    """
    plt.rcParams['text.usetex'] = True
    plt.rcParams['text.usetex'] = False


# Lambda function to calculate dI_phi
dI_phi = lambda b, phase, order: 2 * sum(b[1:order + 1:] * np.sin(2 * np.pi * phase * np.arange(order + 1)[1::]))


def Half_Comp(filter_files, Epoch, period,
              FT_order=10, sections=4, section_order=8,
              resolution=512, offset=0.25, save=False, outName='noname_halfcomp.png',
              title=None, filter_names=None, sans_font=False,
              write_callback=None, cancel_event=None):
    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # Setting font family if not using sans font
    if sans_font == False:
        plt.rcParams['font.family'] = 'serif'  # Set font family to serif if sans_font is False
        plt.rcParams['mathtext.fontset'] = 'dejavuserif'  # Set math font to serif if sans_font is False

    # Calculate the number of bands
    bands = len(filter_files)

    # Create subplots for flux and dI
    axs, _ = plot.multiplot(figsize=(6, 9), dpi=512, height_ratios=[7 / 3 * bands, 3])
    flux = axs[0]  # Flux subplot
    dI = axs[1]  # dI subplot

    colors = ['blue', 'limegreen', 'red', 'm']  # Color list for plots
    styles = ['--', '-.', ':', '--']  # Line styles for plots

    R2_halves = []  # List to store R^2 values for each band

    # Draw horizontal line at y=0 on dI subplot
    dI.axhline(0, linestyle='-', color='black', linewidth=1)

    # Loop over each band
    for band in range(bands):
        # Binning of data
        a, b = binning.polybinner(filter_files[band], Epoch, period, sections=sections,
                                  norm_factor='alt', section_order=section_order)[0][:2:]

        half = int(0.5 * resolution) + 1  # Calculate half index for resolution

        # Compute Fourier Transform for positive and negative bins
        FT1 = FT.FT_plotlist(a, b, FT_order, resolution)
        FT2 = FT.FT_plotlist(a, -1 * b, FT_order, resolution)
        FTphase1, FTflux1 = FT1[0][:half:], FT1[1][:half:]  # Positive phase and flux
        FTphase2, FTflux2 = FT2[0][:half:], FT2[1][:half:]  # Negative phase and flux

        # Calculate coefficient of determination (R^2) between positive and negative fluxes
        R2_halves.append(calc.error.CoD(FTflux1, FTflux2))
        R2_halves.append(calc.error.CoD(FTflux2, FTflux1))

        dIlist = []  # List to store dI values
        for phase in FTphase1:
            dIlist.append(dI_phi(b, phase, FT_order))  # Calculate dI values

        # Adjust flux values for plotting
        FTflux1 = np.array(FTflux1) + (1 - band) * offset
        FTflux2 = np.array(FTflux2) + (1 - band) * offset

        # Plot flux and dI
        flux.plot(FTphase1, FTflux1, linestyle=styles[band], color=colors[band])
        flux.plot(FTphase2, FTflux2, '-', color=colors[band])

        # Add filter names to flux subplot if provided
        if filter_names is not None:
            if len(filter_names) == bands:
                flux.text(-0.12, FTflux1[0], filter_names[band], fontsize=18, rotation=0)

        # Plot dI
        dI.plot(FTphase1, dIlist, linestyle=styles[band], color=colors[band])

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # Set x-axis limit for flux subplot
    plt.xlim(-0.025, 0.525)

    # Set y-axis label for flux subplot if filter_names is None
    if filter_names is None:
        flux.set_ylabel('Flux', fontsize=16)

    # Format subplots
    plot.sm_format(flux, numbersize=15, X=0.125, x=0.025, xbottom=False, bottomspine=False, Y=None)
    plot.sm_format(dI, numbersize=15, X=0.125, x=0.025, xtop=False, topspine=False)
    dI.set_xlabel(r'$\Phi$', fontsize=18)  # Set x-axis label for dI subplot
    dI.set_ylabel(r'$\Delta I(\Phi)_{\rm FT}$', fontsize=18)  # Set y-axis label for dI subplot

    # Set title for flux subplot if title is not empty
    if title != '':
        flux.set_title(title, fontsize=14.4, loc='left')

    # Save figure if save is True
    if save:
        plt.savefig(outName, bbox_inches='tight')
        log(outName + ' saved.')

    # plt.show()  # Show the plot

    # Reset font settings to default
    plt.rcParams['font.family'] = 'sans'
    plt.rcParams['mathtext.fontset'] = 'dejavusans'

    return None  # Return "Done" message


def OConnell_total(inputFile, Epoch, period, order, sims=1000,
                   sections=4, section_order=8, norm_factor='alt',
                   starName='', filterName='', FT_order=10, FTres=500,
                   write_callback=None, cancel_event=None):
    """
    This function calculates various parameters related to the O'Connell Effect.
    It performs Monte Carlo simulations to estimate errors.
    Approximate runtime: ~ sims/1000 minutes.
    """

    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    if cancel_event.is_set():
        log("Task canceled.")
        return

    """
    Generating master parameters from the observational data.
    """
    # ============================== DO NOT CHANGE ==============================
    # Binning the data
    PB = binning.polybinner(inputFile, Epoch, period, sections=sections, norm_factor=norm_factor,
                            section_order=section_order, FT_order=FT_order)
    c_MB = PB[1][0]  # Correct phase bins and fluxes
    nc_MB = PB[1][1]  # Non-corrected phase bins and fluxes
    ob_phaselist = c_MB[0][1]  # Corrected phase list
    ob_fluxlist = c_MB[1][1]  # Corrected flux list
    ob_fluxerr = c_MB[1][2]  # Corrected flux error list
    c_sec_phases = c_MB[5][0]  # Corrected section phases
    nc_sec_phases = nc_MB[5][0]  # Non-corrected section phases
    c_sec_index = c_MB[5][3]  # Corrected section index
    nc_sec_index = nc_MB[5][3]  # Non-corrected section index

    a = PB[0][0]  # Fourier coefficients (cosine terms)
    b = PB[0][1]  # Fourier coefficients (sine terms)
    # ==========================================================================

    FTsynth = FT.synth(a, b, ob_phaselist, FT_order)  # synthetic FT curve
    master_simflux = []
    for sim in tqdm(range(sims), desc='Simulating light curves', position=0):
        master_simflux.append(FT.sim_ob_flux(FTsynth, ob_fluxerr))
    # ============
    if cancel_event.is_set():
        log("Task canceled.")
        return
    # ============
    c_master_simflux = []
    nc_master_simflux = []
    master_polyflux = []
    """
    List of the resampled polynomial fluxes. Each embedded list is different 
    because the generating data has been Monte Carloed.
    """
    master_FTflux = []  # FT fluxes resulting from the simulations
    master_a = []
    master_b = []  # Lists of the a and b FT coefficients for each sim
    OERlist, LCAlist, dIlist, dIavelist = [], [], [], []

    # = begin sim loop =
    for sim in tqdm(range(sims), desc='Simulation processing', position=0):
        """
        Generating empty lists so that stuff can be inserted into the
        specified [sim] index. Otherwise when trying to append list[sim],
        it would throw an error (there'd be nothing to append to).
        """
        c_master_simflux.append([])
        nc_master_simflux.append([])

        # == begin section loop ===
        for section in range(sections):
            c_master_simflux[sim].append([])  # same as before
            nc_master_simflux[sim].append([])

            """
            'Replacing' the observational fluxes with the Monte Carlo fluxes
            based on the index of each data point when it was in ob_fluxlist
            (the MC fluxes are ordered the same as ob_fluxlist).
            """
            for i in range(len(c_sec_phases[section])):
                c_master_simflux[sim][section].append(master_simflux[sim][c_sec_index[section][i]])
            for i in range(len(nc_sec_phases[section])):
                nc_master_simflux[sim][section].append(master_simflux[sim][nc_sec_index[section][i]])
        # == end section loop == ; resume sim loop

        # Generating polynomial resampled fluxes for each simulation.
        minipoly = binning.minipolybinner(c_sec_phases, c_master_simflux[sim],
                                          nc_sec_phases, nc_master_simflux[sim],
                                          section_order)
        master_polyflux.append(minipoly[1])

        # Calculating various parameters for each simulation.
        FTcoef = FT.coefficients(master_polyflux[sim])
        master_a.append(FTcoef[1])
        master_b.append(FTcoef[2])
        master_FTflux.append(FT.FT_plotlist(master_a[sim], master_b[sim], FT_order, FTres)[1])
        OERlist.append(OConnell.OER_FT(master_a[sim], master_b[sim], FT_order))
        LCAlist.append(OConnell.LCA_FT(master_a[sim], master_b[sim], FT_order, 256))
        dIlist.append(OConnell.Delta_I_fixed(master_b[sim], FT_order))
        dIavelist.append(OConnell.Delta_I_mean_obs_noerror(ob_phaselist, ob_fluxlist, phase_range=0.05))
    # = end sim loop =

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # === end Monte Carlo ===

    """
    Calculating FT model errors, as opposed to the errors calculated
    with the simulate light curves.
    """
    a0sig, a0rat = FT.a_sig_fast(a, b, 0, a[0], ob_phaselist, ob_fluxlist, ob_fluxerr, order)
    a_model_err = [a0sig]
    b_model_err = [0]
    a_rat = [a0rat]
    b_rat = [1]
    for n in tqdm(range(1, order + 1), position=0, desc='Calculating FT errors'):
        # TODO, dx0=1 isn't fullproof
        asig, ar = FT.a_sig_fast(a, b, n, a[n], ob_phaselist, ob_fluxlist, ob_fluxerr, order, dx0=1)
        bsig, br = FT.b_sig_fast(a, b, n, b[n], ob_phaselist, ob_fluxlist, ob_fluxerr, order, dx0=1)
        a_model_err.append(asig)
        b_model_err.append(bsig)
        a_rat.append(ar)
        b_rat.append(br)

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # == FT coefficients ==
    a_MC_err = np.array(list(map(st.stdev, zip(*master_a))))
    b_MC_err = np.array(list(map(st.stdev, zip(*master_b))))
    a_total_err = (np.array(a_model_err) ** 2 + a_MC_err[:order + 1:] ** 2) ** 0.5
    b_total_err = (np.array(b_model_err) ** 2 + b_MC_err[:order + 1:] ** 2) ** 0.5

    a2_0125_a2 = a[2] * (0.125 - a[2])
    a2_0125_a2_err = sig_f(lambda x: x * (0.125 - x), a[2], a_total_err[2])

    # == OER ==
    OER = OConnell.OER_FT(a, b, order)
    OER_model_err = OConnell.OER_FT_error_fixed(a, b, a_model_err, b_model_err, order)
    OER_MC_err = st.stdev(OERlist)
    OER_total_err = (OER_model_err ** 2 + OER_MC_err ** 2) ** 0.5

    # == LCA ==
    LCA = OConnell.LCA_FT(a, b, order, 1024)
    LCA_model_err = OConnell.LCA_FT_error(a, b, a_model_err, b_model_err, order, 1024)[1]
    LCA_MC_err = st.stdev(LCAlist)
    LCA_total_err = (LCA_model_err ** 2 + LCA_MC_err ** 2) ** 0.5

    # == Delta_I ==
    Delta_I_025 = OConnell.Delta_I_fixed(b, order)
    Delta_I_025_model_err = OConnell.Delta_I_error_fixed(b_model_err, order)
    Delta_I_025_MC_err = st.stdev(dIlist)
    Delta_I_025_total_err = (Delta_I_025_model_err ** 2 + Delta_I_025_MC_err ** 2) ** 0.5

    # == Delta_I_ave ==
    DIave = OConnell.Delta_I_mean_obs(ob_phaselist, ob_fluxlist, ob_fluxerr, phase_range=0.05, weighted=False)
    Delta_I_ave = DIave[0]
    Delta_I_ave_err = DIave[1]

    # == print parameters ==
    r = lambda x: round(x, 5)

    valerr = lambda x, dx, label, PRECISION=6: print(label + ' =', round(x, PRECISION), '+/-', round(dx, PRECISION))

    # print('\n')
    valerr(a[1], a_total_err[1], 'a1')
    valerr(a[2], a_total_err[2], 'a2')
    valerr(a[4], a_total_err[4], 'a4')
    valerr(a[2] * (0.125 - a[2]),
           sig_f(lambda x: x * (0.125 - x), a[2], (a_model_err[2] ** 2 + a_MC_err[2] ** 2) ** 0.5), 'a2(0.125-a2)')
    # print('')
    valerr(OER, OER_total_err, 'OER')
    # print(r(OER_model_err), r(OER_MC_err), '\n')

    valerr(LCA, LCA_total_err, 'LCA')
    # print(r(LCA_model_err), r(LCA_MC_err), '\n')

    valerr(Delta_I_025, Delta_I_025_total_err, 'Delta_I')
    # print(r(Delta_I_025_model_err), r(Delta_I_025_MC_err), '\n')

    valerr(Delta_I_ave, Delta_I_ave_err, 'Delta_I_ave')

    return [a, a_total_err], [b, b_total_err], [Delta_I_025, Delta_I_025_total_err], [Delta_I_ave, Delta_I_ave_err], [
        OER, OER_total_err], [LCA, LCA_total_err], [a2_0125_a2, a2_0125_a2_err]


def multi_OConnell_total(filter_files, Epoch, period, order=10,
                         sims=1000, sections=4, section_order=8,
                         norm_factor='alt', starName='', filterNames=[r'$\rm B$', r'$\rm V$', r'$\rm R_C$'],
                         FT_order=10, FTres=500, plot_only=False,
                         plotoff=0.25, save=False, outName='noname.png',
                         write_callback=None, cancel_event=None):
    """
    This function generates a half-comparison plot and calculates various parameters related to the O'Connell Effect.
    If plot_only is set to True, only the half-comparison plot will be generated.
    """

    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # Generate half-comparison plot
    Half_Comp(filter_files, Epoch, period, FT_order=order, sections=sections,
              section_order=section_order, offset=plotoff, save=save, outName=outName,
              filter_names=filterNames, write_callback=write_callback, cancel_event=cancel_event)

    if cancel_event.is_set():
        log("Task canceled.")
        return

    # Perform O'Connell analysis
    if not plot_only:
        filters = len(filter_files)
        a_all = []
        a_err_all = []
        b_all = []
        b_err_all = []
        dI_FT = []
        dI_FT_err = []
        dI_ave = []
        dI_ave_err = []
        OERs = []
        OERs_err = []
        LCAs = []
        LCAs_err = []
        a22s = []
        a22s_err = []

        # Running the O'Connell function on each filter
        for band in range(len(filter_files)):
            oc = OConnell_total(filter_files[band], Epoch, period, order, sims=sims,
                                sections=sections, section_order=section_order,
                                norm_factor=norm_factor, FT_order=order, FTres=FTres,
                                filterName='Filter ' + str(band + 1),
                                write_callback=write_callback, cancel_event=cancel_event)
            a_all.append(oc[0][0])
            a_err_all.append(oc[0][1])
            b_all.append(oc[1][0])
            b_err_all.append(oc[1][1])
            dI_FT.append(oc[2][0])
            dI_FT_err.append(oc[2][1])
            dI_ave.append(oc[3][0])
            dI_ave_err.append(oc[3][1])
            OERs.append(oc[4][0])
            OERs_err.append(oc[4][1])
            LCAs.append(oc[5][0])
            LCAs_err.append(oc[5][1])
            a22s.append(oc[6][0])
            a22s_err.append(oc[6][1])

        if cancel_event.is_set():
            log("Task canceled.")
            return

        # LaTeX table creation
        table_header = '\\begin{table}[H]\n' + '\\begin{center}\n' + '\\begin{tabular}{c|'
        for band in range(filters):
            table_header += 'c'
        table_header += '}\n' + '\\hline\\hline\n' + 'Parameter '
        for band in range(filters):
            table_header += '& Filter ' + str(band + 1)
        table_header += '\\\ \n' + '\\hline\n'

        a1_line = '$a_1$ '
        a2_line = '$a_2$ '
        a4_line = '$a_4$ '
        a22_line = '$a_2(0.125-a_2)$ '
        b1_line = '$2b_1$ '
        dIFT_line = '$\Delta I_{\\rm FT}$ '
        dIave_line = '$\Delta I_{\\rm ave}$ '
        OER_line = 'OER '
        LCA_line = 'LCA '

        strr = lambda x, e=5: str(round(x, e))
        for band in range(filters):
            a1_line += '& $' + strr(a_all[band][1]) + '\pm ' + strr(a_err_all[band][1]) + '$ '
            a2_line += '& $' + strr(a_all[band][2]) + '\pm ' + strr(a_err_all[band][2]) + '$ '
            a4_line += '& $' + strr(a_all[band][4]) + '\pm ' + strr(a_err_all[band][4]) + '$ '
            a22_line += '& $' + strr(a22s[band]) + '\pm ' + strr(a22s_err[band]) + '$ '
            b1_line += '& $' + strr(2 * b_all[band][1]) + '\pm ' + strr(2 * b_err_all[band][1]) + '$ '
            dIFT_line += '& $' + strr(dI_FT[band]) + '\pm ' + strr(dI_FT_err[band]) + '$ '
            dIave_line += '& $' + strr(dI_ave[band]) + '\pm ' + strr(dI_ave_err[band]) + '$ '
            OER_line += '& $' + strr(OERs[band]) + '\pm ' + strr(OERs_err[band]) + '$ '
            LCA_line += '& $' + strr(LCAs[band]) + '\pm ' + strr(LCAs_err[band]) + '$ '

        if cancel_event.is_set():
            log("Task canceled.")
            return

        lines = [a1_line, a2_line, a4_line, a22_line, b1_line, dIFT_line, dIave_line, OER_line, LCA_line]
        for count, _ in enumerate(lines):
            lines[count] += '\\\ \n'

        output = table_header
        for _, line in enumerate(lines):
            output += line
        output += '\\hline\n' + '\\end{tabular}\n' + '\\caption{Fourier and O\'Connell stuff (' + str(
            sims) + ' sims)}\n' + '\\label{tbl:OConnell}\n' + '\\end{center}\n' + '\\end{table}\n'
        # End LaTeX table creation

        # print(output)  # Output LaTeX table to console
        # outputfile = input("Please enter an output file name without the extension: ")
        outputfile = outName.rstrip(".pdf")
        file = open(outputfile + ".txt", "w")
        file.write(output)  # Write LaTeX table to file
        file.close()

    return 'nada'



"""
Author: Kyle Koeller
Created: 11/08/2022
Last Edited: 12/16/2024

This program is meant to automatically do the data reduction of the raw images from the
Ball State University Observatory (BSUO) and SARA data. The new calibrated images are placed into a new folder as to
not overwrite the original images.
"""
# import sys
from pathlib import Path
import time
import warnings

from astropy import wcs
from astropy.stats import mad_std
from astropy import units as u
from astropy.io import fits
from astropy.time import Time
from astropy.coordinates import SkyCoord, EarthLocation

import ccdproc as ccdp
import numpy as np

# turn off this warning that just tells the user,
# "The warning raised when the contents of the FITS header have been modified to be standards compliant."
warnings.filterwarnings("ignore", category=wcs.FITSFixedWarning)

# global variables for combining, to either play with or set depending on the run
sigma_clip_low_thresh = None
sigma_clip_high_thresh = 3
sigclip = 5  # sigclip for cosmic ray removal
rdnoise = 10.83  # * u.electron  # gathered from fits headers manually
gain = 1.43  # * u.electron / u.adu  # gathered from fits headers manually
overwrite = True  # if the user wants to overwrite already existing files or not, by default it is set to True
mem_limit = 1600e6  # maximum memory limit 4.5 Gb is the recommended which is 450e6 (i.e. 8.0 Gb would be 800e6)
dark_bool = "True"
location = "bsuo"


def run_reduction(path, calibrated, location, cancel_event, dark_bool=True, overscan_region="[2073:2115, :]", trim_region="[20:2060, 12:2057]", write_callback=None):
    """
    Main function to run the reduction process. It replaces user input with programmatic arguments.

        :param path: The path to the raw images directory
    :param calibrated: The path to the output directory for calibrated images
    :param location: The observing location (e.g., 'bsuo', 'kpno', etc.)
    :param cancel_event:
    :param dark_bool: Boolean to indicate whether to use dark frames
    :param overscan_region: String for the overscan region of the CCD
    :param trim_region: String for the trim region of the CCD
    :param write_callback: Function to write log messages to the GUI
    """
    def log(message):
        """Log messages to the GUI if callback provided, otherwise print"""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    try:
        if cancel_event.is_set():
            log("Task canceled before starting.")
            return

        # Convert paths to Path objects
        images_path = Path(path)
        calibrated_data = Path(calibrated)

        if not images_path.exists():
            raise FileNotFoundError(f"Raw images path '{path}' does not exist.")
        if not calibrated_data.exists():
            calibrated_data.mkdir(parents=True)

        # Configure location-specific defaults
        if location.lower() == "kpno":
            configure_kpno()
        elif location.lower() == "ctio":
            configure_ctio()
        elif location.lower() == "lapalma":
            configure_lapalma()

        # Process files
        files = ccdp.ImageFileCollection(images_path)
        # zero, overscan_region, trim_region = bias(files, calibrated_data, log)
        zero = bias(files, calibrated_data, overscan_region, trim_region, log, cancel_event)
        master_dark = dark(files, zero, calibrated_data, overscan_region, trim_region, log, cancel_event) if dark_bool else None
        flat(files, zero, master_dark, calibrated_data, overscan_region, trim_region, log, cancel_event)
        science_images(files, calibrated_data, zero, master_dark, trim_region, overscan_region, log, cancel_event)

        log("\nReduction process completed successfully.\n")
    except Exception as e:
        log(f"An error occurred: {e}")
        raise


def kpno():
    """
    Kitt Peak National Observatory default values

    :return: None
    """
    global gain, rdnoise, dark_bool

    gain = 2.3
    rdnoise = 6.0
    dark_bool = True


def ctio():
    """
    Cerro Tololo Inter-American Observatory default values

    :return: None
    """
    global gain, rdnoise, dark_bool

    gain = 2.0
    rdnoise = 9.7
    dark_bool = True


def lapalma():
    """
    La Palma default values

    :return: None
    """
    global gain, rdnoise, dark_bool

    gain = 1.0
    rdnoise = 6.3
    dark_bool = True


def new_default():
    """
    Generates new values that the user can enter

    :return: newly entered default values
    """
    global sigma_clip_high_thresh, sigma_clip_low_thresh, gain, rdnoise, sigclip, dark_bool, location

    sigma_clip_low_thresh = (input("\nEnter a sigma clip low threshold, default is 'None': "))
    if sigma_clip_low_thresh.lower() == "none":
        sigma_clip_low_thresh = None
    else:
        sigma_clip_low_thresh = int(sigma_clip_high_thresh)
    sigma_clip_high_thresh = int(input("Enter a sigma clip high threshold, default is '3': "))
    gain = float(input("Enter a gain value, default is '1.43' electron/adu: "))
    rdnoise = float(input("Enter a readnoise value, default is '10.83' electrons: "))
    sigclip = int(input("Enter a sigma clip value for cosmic ray removal, default is '5': "))
    dark_bool = input("Are you using Dark Frames, default is True (enter 'True' or 'False'): ")
    if dark_bool.lower == "false":
        dark_bool = False
    elif dark_bool.lower == "true":
        dark_bool = True
    location = input("What is your observation location, default is bsuo (ctio, kpno, lapalma")


def reduce(ccd, overscan_region, trim_region, num, zero, combined_dark, good_flat):
    """
    This function takes the information for each section of the reduction process into a singular function for
    limits in duplication of the code.

    :param ccd: individual image
    :param overscan_region: the overscan region of the image
    :param trim_region: region of the image to trim
    :param num: tells the program what stage of the process it is in
    :param zero: master bias
    :param combined_dark: master dark
    :param good_flat: master flat
    :return: depends on the stage of process, but will return a final image for each process
    """

    # subtract overscan, trims image, and gain corrects
    if overscan_region.lower() == "none":
        pass
    else:
        # new_ccd = ccdp.ccd_process(ccd, fits_section=overscan_region, trim=trim_region, gain=gain * u.electron / u.adu,
        #                            readnoise=rdnoise * u.electron, error=False)
        ccd = ccdp.subtract_overscan(ccd, fits_section=overscan_region, median=True, overscan_axis=None)

    trim_ccd = ccdp.trim_image(ccd, fits_section=trim_region)
    new_ccd = ccdp.gain_correct(trim_ccd, gain=gain * u.electron / u.adu)  # gain correct the image

    # this if statement checks whether the input is bias, dark, flat, or science reduction
    # bias combining
    if num == 0:
        return new_ccd
    # dark combining
    elif num == 1:
        # print(new_ccd.unit, zero.unit)
        # Subtract bias
        sub_ccd = ccdp.subtract_bias(new_ccd, zero)

        return sub_ccd
    # flat combining
    elif num == 2:
        # print(new_ccd.unit, zero.unit)
        # Subtract bias
        sub_ccd = ccdp.subtract_bias(new_ccd, zero)

        # Subtract the dark current
        if not dark_bool:
            final_ccd = sub_ccd
        else:
            final_ccd = ccdp.subtract_dark(sub_ccd, combined_dark, exposure_time='exptime', exposure_unit=u.second,
                                           scale=True)
        return final_ccd
    # science calibration
    elif num == 3:
        # Subtract bias
        if location.lower() == "bsuo":
            sub_ccd = ccdp.subtract_bias(new_ccd, zero)
        else:
            sub_ccd = ccdp.subtract_bias(new_ccd, zero)
        # Subtract the dark current
        if not dark_bool:
            reduced = sub_ccd
        else:
            reduced = ccdp.subtract_dark(sub_ccd, combined_dark, exposure_time='exptime', exposure_unit=u.second,
                                         scale=True)
        # flat field correct the science image based on filter
        reduced = ccdp.flat_correct(ccd=reduced, flat=good_flat, min_value=1.0)

        # cosmic ray reject above 5 sigmas and gain_apply is set to false because it changes the units of the image
        # new_reduced = ccdp.cosmicray_lacosmic(reduced, gain_apply=False, readnoise=rdnoise, gain=gain, sigclip=sigclip)

        return reduced


def bias(files, calibrated_data, overscan_region, trim_region, log, cancel_event):
    """
    Calibrates bias images and creates a master bias.

    :param files: Image file collection
    :param calibrated_data: Path to save calibrated images
    :param overscan_region: String for the overscan region of the CCD
    :param trim_region: String for the trim region of the CCD
    :param log: Logging function
    :return: Master bias
    """
    log("\nStarting bias calibration.")
    # Simulate overscan and trim region determination
    # overscan_region = "[2073:2115, :]"
    # trim_region = "[20:2060, 12:2057]"

    log(f"Overscan Region: {overscan_region}")
    log(f"Trim Region: {trim_region}")

    for ccd, file_name in files.ccds(imagetyp='BIAS', return_fname=True, ccd_kwargs={'unit': 'adu'}):
        if cancel_event.is_set():
            log("Task canceled.")
            return
        log(f"Processing bias image: {file_name}")
        new_ccd = reduce(ccd, overscan_region, trim_region, 0, zero=None, combined_dark=None, good_flat=None)
        output_path = calibrated_data / f"{file_name.split('.')[0]}.fits"
        new_ccd.write(output_path, overwrite=overwrite)
        log(f"Saved calibrated bias image: {output_path}")

    log("\nCombining bias frames to create master bias.")
    reduced_images = ccdp.ImageFileCollection(calibrated_data)
    calibrated_biases = reduced_images.files_filtered(imagetyp='BIAS', include_path=True)

    combined_bias = ccdp.combine(
        calibrated_biases,
        method='average',
        sigma_clip=True,
        sigma_clip_low_thresh=sigma_clip_low_thresh,
        sigma_clip_high_thresh=sigma_clip_high_thresh,
        sigma_clip_func=np.ma.median,
        mem_limit=mem_limit
    )
    combined_bias.meta['combined'] = True
    combined_bias_path = calibrated_data / "zero.fits"
    combined_bias.write(combined_bias_path, overwrite=overwrite)
    log(f"Master bias created: {combined_bias_path}")

    return combined_bias # , overscan_region, trim_region


def dark(files, zero, calibrated_path, overscan_region, trim_region, log, cancel_event):
    """
    Calibrates dark frames and creates a master dark.

    :param files: Image file collection
    :param zero: Master bias
    :param calibrated_path: Path to save calibrated images
    :param overscan_region: Overscan region for dark frames
    :param trim_region: Trim region for dark frames
    :param log: Logging function
    :return: Master dark frame
    """
    log("\nStarting dark calibration.")
    for ccd, file_name in files.ccds(imagetyp='DARK', return_fname=True, ccd_kwargs={'unit': 'adu'}):
        if cancel_event.is_set():
            log("Task canceled.")
            return

        log(f"Processing dark image: {file_name}")
        sub_ccd = reduce(ccd, overscan_region, trim_region, 1, zero, combined_dark=None, good_flat=None)
        output_path = calibrated_path / f"{file_name.split('.')[0]}.fits"
        sub_ccd.write(output_path, overwrite=overwrite)
        log(f"Saved calibrated dark image: {output_path}")

    log("\nCombining dark frames to create master dark.")
    reduced_images = ccdp.ImageFileCollection(calibrated_path)
    calibrated_darks = reduced_images.files_filtered(imagetyp='DARK', include_path=True)

    combined_dark = ccdp.combine(
        calibrated_darks,
        method='average',
        sigma_clip=True,
        sigma_clip_low_thresh=sigma_clip_low_thresh,
        sigma_clip_high_thresh=sigma_clip_high_thresh,
        sigma_clip_func=np.ma.median,
        mem_limit=mem_limit
    )
    combined_dark.meta['combined'] = True
    combined_dark_path = calibrated_path / "master_dark.fits"
    combined_dark.write(combined_dark_path, overwrite=overwrite)
    log(f"Master dark created: {combined_dark_path}")

    return combined_dark


def flat(files, zero, combined_dark, calibrated_path, overscan_region, trim_region, log, cancel_event):
    """
    Calibrate flat images.

    :param files: file location for raw images
    :param zero: combined bias image
    :param combined_dark: combined bias image
    :param calibrated_path: file location for new images
    :param trim_region: trim region for images
    :param overscan_region: overscan region for images
    :return: master flat files in each filter
    """
    log("\nStarting flat calibration.")

    # calibrating and combining the flat frames
    for ccd, file_name in files.ccds(imagetyp='FLAT', return_fname=True, ccd_kwargs={'unit': 'adu'}):
        if cancel_event.is_set():
            log("Task canceled.")
            return

        final_ccd = reduce(ccd, overscan_region, trim_region, 2, zero, combined_dark, good_flat=None)

        # new file name with the filter and number from the original file
        list_of_words = file_name.split(".")
        new_fname = "{}.fits".format(list_of_words[0])

        # Save the result
        final_ccd.write(calibrated_path / new_fname, overwrite=overwrite)

        add_header(calibrated_path, new_fname, "FLAT", "None", None, None, None, trim_region, overscan_region)

        log("Finished overscan correction, bias subtraction, and dark subtraction for " + str(new_fname))

    log("\nFinished overscan, bias subtracting, and dark subtracting of flat frames.")
    log("\nStarting flat combination.")
    time.sleep(10)

    ifc = ccdp.ImageFileCollection(calibrated_path)
    flat_filters = set(h['FILTER'] for h in ifc.headers(imagetyp="FLAT"))
    for filt in flat_filters:
        to_combine = ifc.files_filtered(imagetyp="flat", filter=filt, include_path=True)
        combined_flats = ccdp.combine(to_combine,
                                      method='median',
                                      sigma_clip_high_thresh=sigma_clip_high_thresh,
                                      sigma_clip_func=np.ma.median, signma_clip_dev_func=mad_std,
                                      rdnoise=rdnoise * u.electron, gain=gain * u.electron / u.adu,
                                      mem_limit=mem_limit
                                      )

        combined_flats.meta['combined'] = True
        flat_file_name = 'master_flat_{}.fits'.format(filt.replace("Empty/", ""))

        combined_flats.write(calibrated_path / flat_file_name, overwrite=overwrite)

        add_header(calibrated_path, flat_file_name, "FLAT", "None", None, None, None, trim_region, overscan_region)

        log("Finished combining flat " + str(flat_file_name))

    log("\nFinished creating the master flats by filter.")


def science_images(files, calibrated_data, zero, combined_dark, trim_region, overscan_region, log, cancel_event):
    all_reds = []
    science_imagetyp = 'LIGHT'
    flat_imagetyp = 'FLAT'

    ifc_reduced = ccdp.ImageFileCollection(calibrated_data)
    combined_flats = {ccd.header['filter']: ccd for ccd in ifc_reduced.ccds(imagetyp=flat_imagetyp, combined=True)}

    log("\nStarting reduction of science images.")

    for light, file_name in files.ccds(imagetyp=science_imagetyp, return_fname=True, ccd_kwargs={'unit': 'adu'}):
        if cancel_event.is_set():
            log("Task canceled.")
            return

        good_flat = combined_flats[light.header['filter']]
        reduced = reduce(light, overscan_region, trim_region, 3, zero, combined_dark, good_flat)

        list_of_words = file_name.split(".")
        new_fname = "{}.fits".format(list_of_words[0])

        all_reds.append(reduced)
        reduced.write(calibrated_data / new_fname, overwrite=overwrite)

        hjd = light.header["JD-HELIO"]
        ra = light.header["RA"]
        dec = light.header["DEC"]

        add_header(calibrated_data, new_fname, science_imagetyp, "None", hjd, ra, dec, trim_region, overscan_region)

        log("Finished calibration of " + str(new_fname))
    log("\nFinished calibrating all science images.")


def add_header(pathway, fname, imagetyp, filter_name, hjd, ra, dec, trim_region, overscan_region):
    """
    Adds values to the header of each image reduced

    :param overscan_region: BIASSEC for the header
    :param trim_region: DATASEC for the header
    :param dec: declination of target object
    :param ra: right ascension of target object
    :param hjd: time of mid-exposure for each image
    :param pathway: pathway to the new file
    :param fname: file name
    :param imagetyp: image type (bias, flat, dark, light)
    :param filter_name: filter type (B, V, R)
    :return: None
    """
    # bias_image = get_pkg_data_filename(pathway + "\\" + fname)
    image_name = pathway / fname
    fits.setval(image_name, "GAIN", value=gain, comment="Units of e-/ADU")
    fits.setval(image_name, "RDNOISE", value=rdnoise, comment="UNits of e-")
    fits.setval(image_name, "OBSERVAT", value=location, comment="Obs. Loc.")
    fits.setval(image_name, "IMAGETYP", value=imagetyp, comment="Image type")
    fits.setval(image_name, "DATASEC", value=trim_region, comment="Trim data section")
    fits.setval(image_name, "BIASSEC", value=overscan_region, comment="Overscan section")
    fits.setval(image_name, "EPOCH", value="J2000.0")
    # fits.setval(bias_image, "FILTER", value=filter_name)
    if imagetyp == "LIGHT":
        if location.lower() == "bsuo":
            # location of BSUO
            obs_location = {
                'lon': -85.411896,  # degrees
                'lat': 40.199879,  # degrees
                'elevation': 0.2873  # kilometers
            }
            bjd = BJD_TDB(hjd, obs_location, ra, dec)
            fits.setval(image_name, "BJD_TDB", value=bjd.value, comment="Bary. Julian Date, Bary. Dynamical Time")
        else:
            bjd = BJD_TDB(hjd, location, ra, dec)
            fits.setval(image_name, "BJD_TDB", value=bjd.value, comment="Bary. Julian Date, Bary. Dynamical Time")


def BJD_TDB(hjd, loc, ra, dec):
    """
    Converts HJD to BJD_TDB

    :param ra: right ascension of target object
    :param dec: declination of target object
    :param hjd: HJD time for the mid-exposure image
    :param loc: location of the observations
    :return: Newly calculated BJD_TDB that is light time corrected
    """
    helio = Time(hjd, scale='utc', format='jd')
    obs = EarthLocation.from_geodetic(loc["lon"], loc["lat"], loc["elevation"])
    star = SkyCoord(ra, dec, unit=(u.hour, u.deg))
    ltt = helio.light_travel_time(star, 'heliocentric', location=obs)
    guess = helio - ltt
    # if we assume guess is correct - how far is heliocentric time away from true value?
    delta = (guess + guess.light_travel_time(star, 'heliocentric', obs)).jd - helio.jd
    # apply this correction
    guess -= delta * u.d

    ltt = guess.light_travel_time(star, 'barycentric', obs)
    return guess.tdb + ltt



"""
Author: John Kielkopf (University of Louisville)
Created: Unknown

Editor: Kyle Koeller
Last Edited: 12/12/2024

Paper is: https://ui.adsabs.harvard.edu/abs/2019ascl.soft05007B/abstract
"""

import numpy as np
import astropy.io.fits as pyfits
from time import gmtime, strftime
from .vseq_updated import conversion
from astropy.time import Time
from astropy.coordinates import EarthLocation, SkyCoord
from astropy import units as u
import os


def process_tess_cutout(search_file, pathway, sector, outprefix, write_callback, cancel_event):
    """
    Process TESS pixel data from a BINTABLE file and save individual images.

    Parameters
    ----------
    search_file : str
        The path to the TESS BINTABLE file.
    pathway : str
        The directory to save output FITS files.
    outprefix : str
        The prefix for the output file names.
    write_callback : function, optional
        A function to log messages to the GUI.
    sector: int
        The sector number being downloaded
    """
    def log(message):
        """Log messages to the GUI or print to the console."""
        if write_callback:
            write_callback(message)
        else:
            print(message)

    try:
        if cancel_event.is_set():
            log(f"Task canceled while processing Sector {sector}.")
            return

        # Extract file name and verify file paths
        infile = "tess" + str(search_file).strip().split("tess")[1]
        fits_path = os.path.join(pathway, infile)

        log(f"Processing file: {infile}")
        log(f"Output directory: {pathway}")

        if not os.path.exists(fits_path):
            raise FileNotFoundError(f"{fits_path} does not exist.")

        # Open FITS file
        inlist = pyfits.open(fits_path)
        inhdr2 = inlist[2].header
        newhdr = inhdr2

        # Remove problematic header fields
        for field in ["INSTRUME", "TELESCOP", "CHECKSUM", "EXTNAME"]:
            newhdr.pop(field, None)

        imagedata = inlist[1].data
        nimages = len(imagedata)

        log("Starting to process images. This may take a few minutes.")
        for i in range(nimages):
            inimage = imagedata[i][4]
            bjd1 = imagedata[i][0]
            quality_flag = imagedata[i][8]
            tess_ffi = imagedata[i][11]

            if np.isnan(bjd1):
                log(f"Image {i + 1} skipped: lacks valid BJD timestamp.")
                continue
            elif quality_flag != 0:
                log(f"Image {i + 1} skipped: poor quality.")
                continue

            # Calculate mid-exposure BJD and HJD
            bjd = 2457000. + bjd1
            split = infile.split("_")
            ra = conversion([str(float(split[1]) / 15)])[0]
            dec = conversion([split[2]])[0]
            hjd = bary_to_helio(ra, dec, bjd, "greenwich")

            # Create output FITS file
            outimage = inimage
            outlist = pyfits.PrimaryHDU(outimage.astype("float32"), newhdr)
            outhdr = outlist.header

            # Update header metadata
            outhdr["LST_UPDT"] = strftime("%Y-%m-%d %H:%M:%S", gmtime())
            outhdr["BJD_TDB"] = bjd
            outhdr["HJD"] = hjd.value
            outhdr["COMMENT"] = tess_ffi

            outfile = f"{pathway}/{outprefix}_tess_{i:05d}.fits"
            outlist.writeto(outfile, overwrite=True)
            log(f"Saved image {i + 1}: {outfile}")

        log("Finished processing all images.")
        inlist.close()
    except Exception as e:
        log(f"An error occurred during processing: {e}")
        raise


def bary_to_helio(ra, dec, bjd, obs_name):
    """Convert BJD to HJD for the given coordinates."""
    bary = Time(bjd, scale="tdb", format="jd")
    obs = EarthLocation.of_site(obs_name)
    star = SkyCoord(ra, dec, unit=(u.hour, u.deg))
    ltt = bary.light_travel_time(star, "barycentric", location=obs)
    guess = bary - ltt
    delta = (guess + guess.light_travel_time(star, "barycentric", obs)).jd - bary.jd
    guess -= delta * u.d
    ltt = guess.light_travel_time(star, "heliocentric", obs)
    return guess.utc + ltt



import unittest
# from unittest import mock
from unittest.mock import patch
from .OC_plot import calculate_oc, arguments


class TestCalculateOC(unittest.TestCase):
    def test_Calculate_OC1(self):
        e, OC, OC_err, _, _ = calculate_oc(10, 0.1, 0, 0.1, 1)
        self.assertEqual(e, 0.5)
        self.assertEqual(OC, '-0.50000')
        self.assertEqual(OC_err, "0.14142")
    
    def test_Calcualte_OC2(self):
        e, OC, OC_err, _, _ = calculate_oc(15.0, 0.1, 5.0, 0.1, 2.0)
        self.assertEqual(e, 5.0)
        self.assertEqual(OC, '0.00000')
        self.assertEqual(OC_err, "0.14142")
    
    def test_Calcualte_OC3(self):
        # line 13 of the test_minimums.txt in the examples folder
        e, OC, OC_err, _, _ = calculate_oc(2458843.932122, 0.00058, 2457143.761819, 0.00014, 0.31297)
        self.assertEqual(e, 5432.5)
        self.assertEqual(OC, '-0.03922')
        self.assertEqual(OC_err, "0.00060")

class TestArguments(unittest.TestCase):
    @patch('builtins.input', side_effect=["10.0", "0.02", "0.312"])
    def test_arguments1(self, mock_inputs):
        result = arguments()
        self.assertEqual(result[0], 10)
        self.assertEqual(result[1], 0.02)
        self.assertEqual(result[2], 0.312)
    
    @patch('builtins.input', side_effect=["100", "0.02", "-0.312"])
    def test_arguments2(self, mock_inputs):
        result = arguments()
        self.assertEqual(result[0], 100.0)
        self.assertEqual(result[1], 0.02)
        self.assertEqual(result[2], -0.312)


unittest.main()
